"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""


"""
This is the interface that will need to be overloaded by the customer so
that his/her code can receive info from the TWS/IBGW.

NOTE: the methods use type annotations to describe the types of the arguments.
This is used by the Decoder to dynamically and automatically decode the
received message into the given EWrapper method. This method can only be
used for the most simple messages, but it's still huge helper.
Also this method currently automatically decode a 'version' field in the
message. However having a 'version' field is a legacy thing, newer
message use the 'unified version': the agreed up min version of both
server and client.

"""

import logging

from .common import * # @UnusedWildImport
from .utils import * # @UnusedWildImport
from .contract import (Contract, ContractDetails, DeltaNeutralContract)
from .order import Order
from .order_state import OrderState
from .execution import Execution
from .ticktype import * # @UnusedWildImport
from .commission_report import CommissionReport


logger = logging.getLogger(__name__)


class EWrapper:
    def __init__(self):
        pass

    def logAnswer(self, fnName, fnParams):
        if logger.isEnabledFor(logging.INFO):
            if 'self' in fnParams:
                prms = dict(fnParams)
                del prms['self']
            else:
                prms = fnParams
            logger.info("ANSWER %s %s", fnName, prms)


    def error(self, reqId:TickerId, errorCode:int, errorString:str):
        """This event is called when there is an error with the
        communication or when TWS wants to send a message to the client."""

        self.logAnswer(current_fn_name(), vars())
        logger.error("ERROR %s %s %s", reqId, errorCode, errorString)


    def winError(self, text:str, lastError:int):
        self.logAnswer(current_fn_name(), vars())


    def connectAck(self):
        """ callback signifying completion of successful connection """
        self.logAnswer(current_fn_name(), vars())


    def marketDataType(self, reqId:TickerId, marketDataType:int):
        """TWS sends a marketDataType(type) callback to the API, where
        type is set to Frozen or RealTime, to announce that market data has been
        switched between frozen and real-time. This notification occurs only
        when market data switches between real-time and frozen. The
        marketDataType( ) callback accepts a reqId parameter and is sent per
        every subscription because different contracts can generally trade on a
        different schedule."""

        self.logAnswer(current_fn_name(), vars())


    def tickPrice(self, reqId:TickerId , tickType:TickType, price:float,
                  attrib:TickAttrib):
        """Market data tick price callback. Handles all price related ticks."""

        self.logAnswer(current_fn_name(), vars())


    def tickSize(self, reqId:TickerId, tickType:TickType, size:Decimal):
        """Market data tick size callback. Handles all size-related ticks."""

        self.logAnswer(current_fn_name(), vars())


    def tickSnapshotEnd(self, reqId:int):
        """When requesting market data snapshots, this market will indicate the
        snapshot reception is finished. """

        self.logAnswer(current_fn_name(), vars())


    def tickGeneric(self, reqId:TickerId, tickType:TickType, value:float):
        self.logAnswer(current_fn_name(), vars())


    def tickString(self, reqId:TickerId, tickType:TickType, value:str):
        self.logAnswer(current_fn_name(), vars())


    def tickEFP(self, reqId:TickerId, tickType:TickType, basisPoints:float,
                formattedBasisPoints:str, totalDividends:float,
                holdDays:int, futureLastTradeDate:str, dividendImpact:float,
                dividendsToLastTradeDate:float):
        self.logAnswer(current_fn_name(), vars())
        """ market data call back for Exchange for Physical
        tickerId -      The request's identifier.
        tickType -      The type of tick being received.
        basisPoints -   Annualized basis points, which is representative of
            the financing rate that can be directly compared to broker rates.
        formattedBasisPoints -  Annualized basis points as a formatted string
            that depicts them in percentage form.
        impliedFuture - The implied Futures price.
        holdDays -  The number of hold days until the lastTradeDate of the EFP.
        futureLastTradeDate -   The expiration date of the single stock future.
        dividendImpact - The dividend impact upon the annualized basis points
            interest rate.
        dividendsToLastTradeDate - The dividends expected until the expiration
            of the single stock future."""

        self.logAnswer(current_fn_name(), vars())


    def orderStatus(self, orderId:OrderId , status:str, filled:Decimal,
                    remaining:Decimal, avgFillPrice:float, permId:int,
                    parentId:int, lastFillPrice:float, clientId:int,
                    whyHeld:str, mktCapPrice: float):
        """This event is called whenever the status of an order changes. It is
        also fired after reconnecting to TWS if the client has any open orders.

        orderId: OrderId - The order ID that was specified previously in the
            call to placeOrder()
        status:str - The order status. Possible values include:
            PendingSubmit - indicates that you have transmitted the order, but have not  yet received confirmation that it has been accepted by the order destination. NOTE: This order status is not sent by TWS and should be explicitly set by the API developer when an order is submitted.
            PendingCancel - indicates that you have sent a request to cancel the order but have not yet received cancel confirmation from the order destination. At this point, your order is not confirmed canceled. You may still receive an execution while your cancellation request is pending. NOTE: This order status is not sent by TWS and should be explicitly set by the API developer when an order is canceled.
            PreSubmitted - indicates that a simulated order type has been accepted by the IB system and that this order has yet to be elected. The order is held in the IB system until the election criteria are met. At that time the order is transmitted to the order destination as specified.
            Submitted - indicates that your order has been accepted at the order destination and is working.
            Cancelled - indicates that the balance of your order has been confirmed canceled by the IB system. This could occur unexpectedly when IB or the destination has rejected your order.
            Filled - indicates that the order has been completely filled.
            Inactive - indicates that the order has been accepted by the system (simulated orders) or an exchange (native orders) but that currently the order is inactive due to system, exchange or other issues.
        filled:int - Specifies the number of shares that have been executed.
            For more information about partial fills, see Order Status for Partial Fills.
        remaining:int -   Specifies the number of shares still outstanding.
        avgFillPrice:float - The average price of the shares that have been executed. This parameter is valid only if the filled parameter value is greater than zero. Otherwise, the price parameter will be zero.
        permId:int -  The TWS id used to identify orders. Remains the same over TWS sessions.
        parentId:int - The order ID of the parent order, used for bracket and auto trailing stop orders.
        lastFilledPrice:float - The last price of the shares that have been executed. This parameter is valid only if the filled parameter value is greater than zero. Otherwise, the price parameter will be zero.
        clientId:int - The ID of the client (or TWS) that placed the order. Note that TWS orders have a fixed clientId and orderId of 0 that distinguishes them from API orders.
        whyHeld:str - This field is used to identify an order held when TWS is trying to locate shares for a short sell. The value used to indicate this is 'locate'.

        """

        self.logAnswer(current_fn_name(), vars())


    def openOrder(self, orderId:OrderId, contract:Contract, order:Order,
                  orderState:OrderState):
        """This function is called to feed in open orders.

        orderID: OrderId - The order ID assigned by TWS. Use to cancel or
            update TWS order.
        contract: Contract - The Contract class attributes describe the contract.
        order: Order - The Order class gives the details of the open order.
        orderState: OrderState - The orderState class includes attributes Used
            for both pre and post trade margin and commission data."""

        self.logAnswer(current_fn_name(), vars())


    def openOrderEnd(self):
        """This is called at the end of a given request for open orders."""

        self.logAnswer(current_fn_name(), vars())


    def connectionClosed(self):
        """This function is called when TWS closes the sockets
        connection with the ActiveX control, or when TWS is shut down."""

        self.logAnswer(current_fn_name(), vars())


    def updateAccountValue(self, key:str, val:str, currency:str,
                            accountName:str):
        """ This function is called only when ReqAccountUpdates on
        EEClientSocket object has been called. """

        self.logAnswer(current_fn_name(), vars())


    def updatePortfolio(self, contract:Contract, position:Decimal,
                        marketPrice:float, marketValue:float,
                        averageCost:float, unrealizedPNL:float,
                        realizedPNL:float, accountName:str):
        """This function is called only when reqAccountUpdates on
        EEClientSocket object has been called."""

        self.logAnswer(current_fn_name(), vars())


    def updateAccountTime(self, timeStamp:str):
        self.logAnswer(current_fn_name(), vars())


    def accountDownloadEnd(self, accountName:str):
        """This is called after a batch updateAccountValue() and
        updatePortfolio() is sent."""

        self.logAnswer(current_fn_name(), vars())


    def nextValidId(self, orderId:int):
        """ Receives next valid order id."""

        self.logAnswer(current_fn_name(), vars())


    def contractDetails(self, reqId:int, contractDetails:ContractDetails):
        """Receives the full contract's definitions. This method will return all
        contracts matching the requested via EEClientSocket::reqContractDetails.
        For example, one can obtain the whole option chain with it."""

        self.logAnswer(current_fn_name(), vars())


    def bondContractDetails(self, reqId:int, contractDetails:ContractDetails):
        """This function is called when reqContractDetails function
        has been called for bonds."""

        self.logAnswer(current_fn_name(), vars())


    def contractDetailsEnd(self, reqId:int):
        """This function is called once all contract details for a given
        request are received. This helps to define the end of an option
        chain."""

        self.logAnswer(current_fn_name(), vars())


    def execDetails(self, reqId:int, contract:Contract, execution:Execution):
        """This event is fired when the reqExecutions() functions is
        invoked, or when an order is filled.  """

        self.logAnswer(current_fn_name(), vars())


    def execDetailsEnd(self, reqId:int):
        """This function is called once all executions have been sent to
        a client in response to reqExecutions()."""

        self.logAnswer(current_fn_name(), vars())



    def updateMktDepth(self, reqId:TickerId , position:int, operation:int,
                        side:int, price:float, size:Decimal):
        """Returns the order book.

        tickerId -  the request's identifier
        position -  the order book's row being updated
        operation - how to refresh the row:
            0 = insert (insert this new order into the row identified by 'position')
            1 = update (update the existing order in the row identified by 'position')
            2 = delete (delete the existing order at the row identified by 'position').
        side -  0 for ask, 1 for bid
        price - the order's price
        size -  the order's size"""

        self.logAnswer(current_fn_name(), vars())


    def updateMktDepthL2(self, reqId:TickerId , position:int, marketMaker:str,
                          operation:int, side:int, price:float, size:Decimal, isSmartDepth:bool):
        """Returns the order book.

        tickerId -  the request's identifier
        position -  the order book's row being updated
        marketMaker - the exchange holding the order
        operation - how to refresh the row:
            0 = insert (insert this new order into the row identified by 'position')
            1 = update (update the existing order in the row identified by 'position')
            2 = delete (delete the existing order at the row identified by 'position').
        side -  0 for ask, 1 for bid
        price - the order's price
        size -  the order's size
        isSmartDepth - is SMART Depth request"""

        self.logAnswer(current_fn_name(), vars())


    def updateNewsBulletin(self, msgId:int, msgType:int, newsMessage:str,
                           originExch:str):
        """ provides IB's bulletins
        msgId - the bulletin's identifier
        msgType - one of: 1 - Regular news bulletin 2 - Exchange no longer
            available for trading 3 - Exchange is available for trading
        message - the message
        origExchange -    the exchange where the message comes from.  """

        self.logAnswer(current_fn_name(), vars())


    def managedAccounts(self, accountsList:str):
        """Receives a comma-separated string with the managed account ids."""
        self.logAnswer(current_fn_name(), vars())


    def receiveFA(self, faData:FaDataType , cxml:str):
        """ receives the Financial Advisor's configuration available in the TWS

        faDataType - one of:
            Groups: offer traders a way to create a group of accounts and apply
                 a single allocation method to all accounts in the group.
            Profiles: let you allocate shares on an account-by-account basis
                using a predefined calculation value.
            Account Aliases: let you easily identify the accounts by meaningful
                 names rather than account numbers.
        faXmlData -  the xml-formatted configuration """

        self.logAnswer(current_fn_name(), vars())

    def historicalData(self, reqId: int, bar: BarData):
        """ returns the requested historical data bars

        reqId - the request's identifier
        date  - the bar's date and time (either as a yyyymmss hh:mm:ssformatted
             string or as system time according to the request)
        open  - the bar's open point
        high  - the bar's high point
        low   - the bar's low point
        close - the bar's closing point
        volume - the bar's traded volume if available
        count - the number of trades during the bar's timespan (only available
            for TRADES).
        WAP -   the bar's Weighted Average Price
        hasGaps  -indicates if the data has gaps or not. """

        self.logAnswer(current_fn_name(), vars())


    def historicalDataEnd(self, reqId:int, start:str, end:str):
        """ Marks the ending of the historical bars reception. """
        self.logAnswer(current_fn_name(), vars())


    def scannerParameters(self, xml:str):
        """ Provides the xml-formatted parameters available to create a market
        scanner.

        xml -   the xml-formatted string with the available parameters."""
        self.logAnswer(current_fn_name(), vars())


    def scannerData(self, reqId:int, rank:int, contractDetails:ContractDetails,
                     distance:str, benchmark:str, projection:str, legsStr:str):
        """ Provides the data resulting from the market scanner request.

        reqid - the request's identifier.
        rank -  the ranking within the response of this bar.
        contractDetails - the data's ContractDetails
        distance -      according to query.
        benchmark -     according to query.
        projection -    according to query.
        legStr - describes the combo legs when the scanner is returning EFP"""

        self.logAnswer(current_fn_name(), vars())


    def scannerDataEnd(self, reqId:int):
        """ Indicates the scanner data reception has terminated.

        reqId - the request's identifier"""

        self.logAnswer(current_fn_name(), vars())


    def realtimeBar(self, reqId: TickerId, time:int, open_: float, high: float, low: float, close: float,
                        volume: Decimal, wap: Decimal, count: int):

        """ Updates the real time 5 seconds bars

        reqId - the request's identifier
        bar.time  - start of bar in unix (or 'epoch') time
        bar.endTime - for synthetic bars, the end time (requires TWS v964). Otherwise -1.
        bar.open_  - the bar's open value
        bar.high  - the bar's high value
        bar.low   - the bar's low value
        bar.close - the bar's closing value
        bar.volume - the bar's traded volume if available
        bar.WAP   - the bar's Weighted Average Price
        bar.count - the number of trades during the bar's timespan (only available
            for TRADES)."""

        self.logAnswer(current_fn_name(), vars())

    def currentTime(self, time:int):
        """ Server's current time. This method will receive IB server's system
        time resulting after the invokation of reqCurrentTime. """

        self.logAnswer(current_fn_name(), vars())


    def fundamentalData(self, reqId:TickerId , data:str):
        """This function is called to receive fundamental
        market data. The appropriate market data subscription must be set
        up in Account Management before you can receive this data."""

        self.logAnswer(current_fn_name(), vars())


    def deltaNeutralValidation(self, reqId:int, deltaNeutralContract:DeltaNeutralContract):
        """Upon accepting a Delta-Neutral RFQ(request for quote), the
        server sends a deltaNeutralValidation() message with the DeltaNeutralContract
        structure. If the delta and price fields are empty in the original
        request, the confirmation will contain the current values from the
        server. These values are locked when the RFQ is processed and remain
        locked until the RFQ is canceled."""

        self.logAnswer(current_fn_name(), vars())



    def commissionReport(self, commissionReport:CommissionReport):
        """The commissionReport() callback is triggered as follows:
        - immediately after a trade execution
        - by calling reqExecutions()."""

        self.logAnswer(current_fn_name(), vars())


    def position(self, account:str, contract:Contract, position:Decimal,
                 avgCost:float):
        """This event returns real-time positions for all accounts in
        response to the reqPositions() method."""

        self.logAnswer(current_fn_name(), vars())


    def positionEnd(self):
        """This is called once all position data for a given request are
        received and functions as an end marker for the position() data. """

        self.logAnswer(current_fn_name(), vars())


    def accountSummary(self, reqId:int, account:str, tag:str, value:str,
                       currency:str):
        """Returns the data from the TWS Account Window Summary tab in
        response to reqAccountSummary()."""

        self.logAnswer(current_fn_name(), vars())


    def accountSummaryEnd(self, reqId:int):
        """This method is called once all account summary data for a
        given request are received."""

        self.logAnswer(current_fn_name(), vars())


    def verifyMessageAPI(self, apiData:str):
        """ Deprecated Function """
        self.logAnswer(current_fn_name(), vars())


    def verifyCompleted(self, isSuccessful:bool, errorText:str):

        self.logAnswer(current_fn_name(), vars())


    def verifyAndAuthMessageAPI(self, apiData:str, xyzChallange:str):

        self.logAnswer(current_fn_name(), vars())


    def verifyAndAuthCompleted(self, isSuccessful:bool, errorText:str):

        self.logAnswer(current_fn_name(), vars())


    def displayGroupList(self, reqId:int, groups:str):
        """This callback is a one-time response to queryDisplayGroups().

        reqId - The requestId specified in queryDisplayGroups().
        groups - A list of integers representing visible group ID separated by
            the | character, and sorted by most used group first. This list will
             not change during TWS session (in other words, user cannot add a
            new group; sorting can change though)."""

        self.logAnswer(current_fn_name(), vars())


    def displayGroupUpdated(self, reqId:int, contractInfo:str):
        """This is sent by TWS to the API client once after receiving
        the subscription request subscribeToGroupEvents(), and will be sent
        again if the selected contract in the subscribed display group has
        changed.

        requestId - The requestId specified in subscribeToGroupEvents().
        contractInfo - The encoded value that uniquely represents the contract
            in IB. Possible values include:
            none = empty selection
            contractID@exchange = any non-combination contract.
                Examples: 8314@SMART for IBM SMART; 8314@ARCA for IBM @ARCA.
            combo = if any combo is selected.  """

        self.logAnswer(current_fn_name(), vars())


    def positionMulti(self, reqId:int, account:str, modelCode:str,
                      contract:Contract, pos:Decimal, avgCost:float):
        """same as position() except it can be for a certain
        account/model"""

        self.logAnswer(current_fn_name(), vars())


    def positionMultiEnd(self, reqId:int):
        """same as positionEnd() except it can be for a certain
        account/model"""

        self.logAnswer(current_fn_name(), vars())


    def accountUpdateMulti(self, reqId:int, account:str, modelCode:str,
                            key:str, value:str, currency:str):
        """same as updateAccountValue() except it can be for a certain
        account/model"""

        self.logAnswer(current_fn_name(), vars())


    def accountUpdateMultiEnd(self, reqId:int):
        """same as accountDownloadEnd() except it can be for a certain
        account/model"""

        self.logAnswer(current_fn_name(), vars())


    def tickOptionComputation(self, reqId:TickerId, tickType:TickType, tickAttrib:int,
            impliedVol:float, delta:float, optPrice:float, pvDividend:float,
            gamma:float, vega:float, theta:float, undPrice:float):
        """This function is called when the market in an option or its
        underlier moves. TWS's option model volatilities, prices, and
        deltas, along with the present value of dividends expected on that
        options underlier are received."""

        self.logAnswer(current_fn_name(), vars())


    def securityDefinitionOptionParameter(self, reqId:int, exchange:str,
        underlyingConId:int, tradingClass:str, multiplier:str,
        expirations:SetOfString, strikes:SetOfFloat):
        """ Returns the option chain for an underlying on an exchange
        specified in reqSecDefOptParams There will be multiple callbacks to
        securityDefinitionOptionParameter if multiple exchanges are specified
        in reqSecDefOptParams

        reqId - ID of the request initiating the callback
        underlyingConId - The conID of the underlying security
        tradingClass -  the option trading class
        multiplier -    the option multiplier
        expirations - a list of the expiries for the options of this underlying
             on this exchange
        strikes - a list of the possible strikes for options of this underlying
             on this exchange """

        self.logAnswer(current_fn_name(), vars())


    def securityDefinitionOptionParameterEnd(self, reqId:int):
        """ Called when all callbacks to securityDefinitionOptionParameter are
        complete

        reqId - the ID used in the call to securityDefinitionOptionParameter """

        self.logAnswer(current_fn_name(), vars())


    def softDollarTiers(self, reqId:int, tiers:list):
        """ Called when receives Soft Dollar Tier configuration information

        reqId - The request ID used in the call to EEClient::reqSoftDollarTiers
        tiers - Stores a list of SoftDollarTier that contains all Soft Dollar
            Tiers information """

        self.logAnswer(current_fn_name(), vars())


    def familyCodes(self, familyCodes:ListOfFamilyCode):
        """ returns array of family codes """
        self.logAnswer(current_fn_name(), vars())


    def symbolSamples(self, reqId:int,
                      contractDescriptions:ListOfContractDescription):
        """ returns array of sample contract descriptions """
        self.logAnswer(current_fn_name(), vars())

    def mktDepthExchanges(self, depthMktDataDescriptions:ListOfDepthExchanges):
        """ returns array of exchanges which return depth to UpdateMktDepthL2"""
        self.logAnswer(current_fn_name(), vars())

    def tickNews(self, tickerId: int, timeStamp:int, providerCode:str, articleId:str, headline:str, extraData:str):
        """ returns news headlines"""
        self.logAnswer(current_fn_name(), vars())

    def smartComponents(self, reqId:int, smartComponentMap:SmartComponentMap):
        """returns exchange component mapping"""
        self.logAnswer(current_fn_name(), vars())

    def tickReqParams(self, tickerId:int, minTick:float, bboExchange:str, snapshotPermissions:int):
        """returns exchange map of a particular contract"""
        self.logAnswer(current_fn_name(), vars())

    def newsProviders(self, newsProviders:ListOfNewsProviders):
        """returns available, subscribed API news providers"""
        self.logAnswer(current_fn_name(), vars())

    def newsArticle(self, requestId:int, articleType:int, articleText:str):
        """returns body of news article"""
        self.logAnswer(current_fn_name(), vars())

    def historicalNews(self, requestId:int, time:str, providerCode:str, articleId:str, headline:str):
        """returns historical news headlines"""
        self.logAnswer(current_fn_name(), vars())

    def historicalNewsEnd(self, requestId:int, hasMore:bool):
        """signals end of historical news"""
        self.logAnswer(current_fn_name(), vars())

    def headTimestamp(self, reqId:int, headTimestamp:str):
        """returns earliest available data of a type of data for a particular contract"""
        self.logAnswer(current_fn_name(), vars())

    def histogramData(self, reqId:int, items:HistogramData):
        """returns histogram data for a contract"""
        self.logAnswer(current_fn_name(), vars())

    def historicalDataUpdate(self, reqId: int, bar: BarData):
        """returns updates in real time when keepUpToDate is set to True"""
        self.logAnswer(current_fn_name(), vars())

    def rerouteMktDataReq(self, reqId: int, conId: int, exchange: str):
        """returns reroute CFD contract information for market data request"""
        self.logAnswer(current_fn_name(), vars())

    def rerouteMktDepthReq(self, reqId: int, conId: int, exchange: str):
        """returns reroute CFD contract information for market depth request"""
        self.logAnswer(current_fn_name(), vars())

    def marketRule(self, marketRuleId: int, priceIncrements: ListOfPriceIncrements):
        """returns minimum price increment structure for a particular market rule ID"""
        self.logAnswer(current_fn_name(), vars())

    def pnl(self, reqId: int, dailyPnL: float, unrealizedPnL: float, realizedPnL: float):
        """returns the daily PnL for the account"""
        self.logAnswer(current_fn_name(), vars())

    def pnlSingle(self, reqId: int, pos: Decimal, dailyPnL: float, unrealizedPnL: float, realizedPnL: float, value: float):
        """returns the daily PnL for a single position in the account"""
        self.logAnswer(current_fn_name(), vars())

    def historicalTicks(self, reqId: int, ticks: ListOfHistoricalTick, done: bool):
        """returns historical tick data when whatToShow=MIDPOINT"""
        self.logAnswer(current_fn_name(), vars())

    def historicalTicksBidAsk(self, reqId: int, ticks: ListOfHistoricalTickBidAsk, done: bool):
        """returns historical tick data when whatToShow=BID_ASK"""
        self.logAnswer(current_fn_name(), vars())

    def historicalTicksLast(self, reqId: int, ticks: ListOfHistoricalTickLast, done: bool):
        """returns historical tick data when whatToShow=TRADES"""
        self.logAnswer(current_fn_name(), vars())

    def tickByTickAllLast(self, reqId: int, tickType: int, time: int, price: float,
                          size: Decimal, tickAttribLast: TickAttribLast, exchange: str,
                          specialConditions: str):
        """returns tick-by-tick data for tickType = "Last" or "AllLast" """
        self.logAnswer(current_fn_name(), vars())

    def tickByTickBidAsk(self, reqId: int, time: int, bidPrice: float, askPrice: float,
                         bidSize: Decimal, askSize: Decimal, tickAttribBidAsk: TickAttribBidAsk):
        """returns tick-by-tick data for tickType = "BidAsk" """
        self.logAnswer(current_fn_name(), vars())

    def tickByTickMidPoint(self, reqId: int, time: int, midPoint: float):
        """returns tick-by-tick data for tickType = "MidPoint" """
        self.logAnswer(current_fn_name(), vars())

    def orderBound(self, reqId: int, apiClientId: int, apiOrderId: int):
        """returns orderBound notification"""
        self.logAnswer(current_fn_name(), vars())
        
    def completedOrder(self, contract:Contract, order:Order, orderState:OrderState):
        """This function is called to feed in completed orders.

        contract: Contract - The Contract class attributes describe the contract.
        order: Order - The Order class gives the details of the completed order.
        orderState: OrderState - The orderState class includes completed order status details."""

        self.logAnswer(current_fn_name(), vars())

    def completedOrdersEnd(self):
        """This is called at the end of a given request for completed orders."""

        self.logAnswer(current_fn_name(), vars())

    def replaceFAEnd(self, reqId: int, text: str):
        """This is called at the end of a replace FA."""

        self.logAnswer(current_fn_name(), vars())

    def wshMetaData(self, reqId: int, dataJson: str):
        self.logAnswer(current_fn_name(), vars())

    def wshEventData(self, reqId: int, dataJson: str):
        self.logAnswer(current_fn_name(), vars())


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Nov 13 19:51:56 2021
@author: ron
"""

from .base import AbstractStrategy
from ..event.event import SignalEvent, OrderType, Direction
from ..common import get_cur_time

class BuyAndHold(AbstractStrategy):
    """
    session_type: 'backtest' or 'live'
    """
    
    def __init__(self, event_queue, session_type, long_only=True, 
                 stoploss=None, portfolio_handler=None, db_client=None
        ):
        
        self.db_client = db_client
        self.tickers_data = {}
        self.event_queue = event_queue
        self.invested = False
        self.long_only = long_only
        self.session_type = session_type
        
        if stoploss is not None and portfolio_handler is None:
            raise ValueError("portfolio_handler must be set in order to calculate stop loss")
        else:
            self.stoploss = stoploss
            self.portfolio_handler = portfolio_handler

    
    def calculate_signal(self, bar_event):           
            if not self.invested:                    
                cur_time = get_cur_time(bar_event, self.session_type)
                signal_event = SignalEvent(cur_time, bar_event.ticker,
                                           OrderType.MARKET, Direction.LONG, 
                                           bar_event.close_price)
                print(signal_event)
                self.event_queue.put(signal_event)
                self.invested = True
                
                
    def get_ticker(self):
        return self.tickers_data

    
    def __str__(self):
        return "Buy And Hold Strategy"


from .base import AbstractOrderHandler
from ..event.event import OrderEvent
from ..common import get_cur_time


class FixedOrderHandler(AbstractOrderHandler):
    """
    TO-DO: Reject the order when cash is not enough 
    """
    def __init__(self, quantity, portfolio, event_queue, session_type, db_client=None):
        self.db_client = db_client
        self.fixed_quantity = quantity
        self.portfolio = portfolio
        self.event_queue = event_queue
        self.session_type = session_type

    
    def create_order(self, signal_event):
        cur_time = get_cur_time(signal_event, self.session_type)
        ticker = signal_event.ticker
        order_type = signal_event.order_type
        direction = signal_event.direction
        entry_price = signal_event.entry_price
        order_event = OrderEvent(cur_time, ticker, order_type, 
                                 direction, self.fixed_quantity, entry_price)
        self.event_queue.put(order_event)
        print(order_event)





"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""


from .object_implem import Object
from .common import UNSET_DECIMAL
from .utils import decimalMaxString


class Execution(Object):
    def __init__(self):
        self.execId = ""
        self.time =  ""
        self.acctNumber =  ""
        self.exchange =  ""
        self.side = ""
        self.shares = UNSET_DECIMAL
        self.price = 0. 
        self.permId = 0
        self.clientId = 0
        self.orderId = 0
        self.liquidation = 0
        self.cumQty = UNSET_DECIMAL
        self.avgPrice = 0.
        self.orderRef =  ""
        self.evRule =  ""
        self.evMultiplier = 0.
        self.modelCode =  ""
        self.lastLiquidity = 0

    def __str__(self):
        return "ExecId: %s, Time: %s, Account: %s, Exchange: %s, Side: %s, Shares: %s, Price: %f, PermId: %d, " \
                "ClientId: %d, OrderId: %d, Liquidation: %d, CumQty: %s, AvgPrice: %f, OrderRef: %s, EvRule: %s, " \
                "EvMultiplier: %f, ModelCode: %s, LastLiquidity: %d" % (self.execId, self.time, self.acctNumber, 
                self.exchange, self.side, decimalMaxString(self.shares), self.price, self.permId, self.clientId, self.orderId, self.liquidation,
                decimalMaxString(self.cumQty), self.avgPrice, self.orderRef, self.evRule, self.evMultiplier, self.modelCode, self.lastLiquidity)


class ExecutionFilter(Object):
    # Filter fields
    def __init__(self):
        self.clientId = 0
        self.acctCode = ""
        self.time = ""
        self.symbol = ""
        self.secType = ""
        self.exchange = "" 
        self.side = ""




from abc import ABC, abstractmethod
from datetime import datetime, timezone

class AbstractOrderHandler(ABC):
    """
    AbstractOrderHandler is an abstract base class 
    for all order handler class

    It receives SignalEvent and then create OrderEvent
    """
    @abstractmethod
    def create_order(self, signal_event):
        pass
    

    def _save_to_db(self, order_event):
        query = ("INSERT INTO OrderEvent (Datetime, Ticker, Order_Type, Direction, Quantity, Entry_Price)"
                "VALUES (%s, %s, %s, %s, %s, %s)"
        )
        timestamp_converted = datetime.fromtimestamp(order_event.timestamp, timezone.utc)
        data = (timestamp_converted, order_event.ticker, order_event.order_type.name, order_event.direction.name, 
                order_event.quantity, order_event.entry_price)
        cursor = self.db_client.cursor()
        cursor.execute(query, data)
        self.db_client.commit()  

    
    
    
    

    
    



"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""

""" Package implementing the Python API for the TWS/IB Gateway """

VERSION = {
    'major': 10,
    'minor': 11,
    'micro': 1}


def get_version_string():
    version = '{major}.{minor}.{micro}'.format(**VERSION)
    return version

__version__ = get_version_string()



from abc import ABC, abstractmethod
from datetime import datetime, timezone


class AbstractDataHandler(ABC):
    """
    AbstractDataHandler is an abstract base class for all 
    live and historical data handlers (e.g. Bar, Tick, Options, Futures...)
    """   
    @abstractmethod
    def get_next(self):
        pass
    
    
    def save_to_db(self, bar_event):
        query = ("INSERT INTO BarEvent (Datetime, Ticker, Open, High, Low, Close, Volume)"
                "VALUES (%s, %s, %s, %s, %s, %s, %s)"
        )
        timestamp_converted = datetime.fromtimestamp(bar_event.timestamp, timezone.utc)
        data = (timestamp_converted, bar_event.ticker, bar_event.open_price, bar_event.high_price, 
                bar_event.low_price, bar_event.close_price, bar_event.volume)
        cursor = self.db_client.cursor()
        cursor.execute(query, data)
        self.db_client.commit()   
    
    




from datetime import datetime, date, timedelta, time, timezone
import calendar
from .base import BaseSchedule

class USStockSchedule(BaseSchedule):
    def __init__(self, year=datetime.utcnow().year):
        super().__init__()
        self.close_early_date = set()
        self.summer_start_date = None
        self.winter_start_date = None
        self._config(year)


    def get_trading_hours(self, date):
        #Date: a date object
        #winter trading time: 1430 - 2100
        #summer trading time: 1330 - 2000
        #day before independence day 1330 - 1700
        #black friday 1430 - 1800
        #christmas eve 1430 - 1800
        #return timestamp are in UTC
        is_summer_time = self.summer_start_date <= date < self.winter_start_date
        
        if is_summer_time:
            start = datetime.combine(date, time(13, 30, tzinfo=timezone.utc)).timestamp()
            if date in self.close_early_date:
                end = datetime.combine(date, time(17, 00, tzinfo=timezone.utc)).timestamp()
            else:
                end = datetime.combine(date, time(20, 00, tzinfo=timezone.utc)).timestamp()
            return (int(start), int(end))
        #winter time   
        else:
            start = datetime.combine(date, time(14, 30, tzinfo=timezone.utc)).timestamp()
            if date in self.close_early_date:
                end = datetime.combine(date, time(18, 00, tzinfo=timezone.utc)).timestamp()
            else:
                end = datetime.combine(date, time(21, 00, tzinfo=timezone.utc)).timestamp()
            return (int(start), int(end))

  
    def _config(self, year):
        #Winter time begins on the first Sunday of Nov
        #Summer time begins on the second Sunday of Mar
        self.winter_start_date = self._find_nth_day(year, 11, 6, 1)
        self.summer_start_date = self._find_nth_day(year, 3, 6, 2)
        self._gen_federal_holidays(year)
        self._gen_weekends(year)
        

    def _gen_federal_holidays(self, year):
        #generate holidays of current year
        #and days with market closed early
        #and start date of summer and winter time 

        #New Year's Day (possibly moved to Monday if on Sunday)
        #If it is on Saturday, it will not precede to Friday (NYSE Rule 7.2)
        holiday = date(year, 1, 1)
        if holiday.isoweekday() == 7:
            holiday += timedelta(days=1)            
        if holiday.isoweekday() != 6:
            self.calendar.add_holiday(holiday)

        #Martin Luther King's birthday (third Monday in January)
        holiday = self._find_nth_day(year, 1, 0, 3)
        self.calendar.add_holiday(holiday)

        #Washington's birthday (third Monday in February)
        holiday = self._find_nth_day(year, 2, 0, 3)
        self.calendar.add_holiday(holiday)

        #Good Friday (Friday before easter)
        holiday = self._calc_easter(year)
        while holiday.isoweekday() != 5:
             holiday -= timedelta(days=1)
        self.calendar.add_holiday(holiday)

        #Memorial Day (last Monday in May)
        total_monday = len([1 for i in calendar.monthcalendar(year, 5) if i[0] != 0])
        holiday = self._find_nth_day(year, 5, 0, total_monday)
        self.calendar.add_holiday(holiday)

        #Juneteenth (Monday if Sunday or Friday if Saturday)
        holiday = date(year, 6, 19)
        if holiday.isoweekday() == 7:
            holiday += timedelta(days=1)
        elif holiday.isoweekday() == 6:
            holiday -= timedelta(days=1)
        self.calendar.add_holiday(holiday)

        #Independence Day (Monday if Sunday or Friday if Saturday)
        #market is closed early on the day before Independence Day 
        holiday = date(year, 7, 4)
        if holiday.isoweekday() == 7:
            holiday += timedelta(days=1)
        elif holiday.isoweekday() == 6:
            holiday -= timedelta(days=1)
        self.calendar.add_holiday(holiday)       
        if 2 <= holiday.isoweekday() <= 5:      
            date_close_early = holiday - timedelta(days=1)
            self.close_early_date.add(date_close_early)

        #Labor Day (first Monday in September)
        holiday = self._find_nth_day(year, 9, 0, 1)
        self.calendar.add_holiday(holiday)

        #Thanksgiving Day (fourth Thursday in November)
        #market is closed early on the day after Thanksgiving
        holiday = self._find_nth_day(year, 11, 3, 4)
        date_close_early = holiday + timedelta(days=1)
        self.close_early_date.add(date_close_early)
        self.calendar.add_holiday(holiday)

        #Christmas (Monday if Sunday or Friday if Saturday)
        #if Christams is on 12/24, the market should be closed on that day rather than closed early 
        holiday = date(year, 12, 25)
        date_close_early = holiday - timedelta(days=1)
        #market is not closed early if Christams is on Mon, Sat, or Sun
        if 2 <= holiday.isoweekday() <= 5:
            self.close_early_date.add(date_close_early)
        if holiday.isoweekday() == 7:
            holiday += timedelta(days=1)
        elif holiday.isoweekday() == 6:
            holiday -= timedelta(days=1)
        self.calendar.add_holiday(holiday)

                   
    def _find_nth_day(self, year, month, day, nth):
        #day: Monday = 0 to Sunday = 6
        if date(year, month, 1).weekday() == day:
            nth -= 1        
        return calendar.Calendar(day).monthdatescalendar(year, month)[nth][0]


    def _calc_easter(self, year):
        #Use Butcher's Algorithm to calculate Easter
        a = year % 19
        b = year // 100
        c = year % 100
        d = (19 * a + b - b // 4 - ((b - (b + 8) // 25 + 1) // 3) + 15) % 30
        e = (32 + 2 * (b % 4) + 2 * (c // 4) - d - (c % 4)) % 7
        f = d + e - 7 * ((a + 11 * d + 22 * e) // 451) + 114
        month = f // 31
        day = f % 31 + 1    
        return date(year, month, day)



    




import pandas as pd
import numpy as np
from math import sqrt
import matplotlib.pyplot as plt


class Statistics:
    """
    Simple Statistics provides a bare-bones example of statistics
    that can be collected through trading.
    Statistics included are Sharpe Ratio, Drawdown, Max Drawdown,
    Max Drawdown Duration.
    TODO need some kind of trading-frequency parameter in setup.
    Sharpe calculations need to know if daily, hourly, minutely, etc.
    """
    def __init__(self, init_asset_value):
        self.init_asset_value = init_asset_value
        self.drawdowns = []
        self.timeseries = []
        # Initialize in order for first-step calculations to be correct.
        self.hwm = [init_asset_value]
        self.asset_value = []
        #self.bars = []
        self.returns = []


    def update(self, timestamp, asset_value):
        """
        Update all statistics that must be tracked over time.
        """
        self.timeseries.append(timestamp)
        self.asset_value.append(asset_value)
        # Calculate Drawdown.
        self.hwm.append(max(self.hwm[-1], self.asset_value[-1]))
        self.drawdowns.append((self.asset_value[-1] - self.hwm[-1])/self.hwm[-1])
        
        if len(self.asset_value) > 1:
            pct = ((self.asset_value[-1] - self.asset_value[-2]) / self.asset_value[-2])
            self.returns.append(pct)


    def get_results(self):
        """
        Return a dict with all important results & stats.
        """
        statistics = {}
        statistics["return"] = ((self.asset_value[-1] - self.init_asset_value) 
                               / self.init_asset_value) * 100
        statistics["sharpe_ratio"] = self.calculate_sharpe_ratio(0.01)
        statistics["max_drawdown"] = min(self.drawdowns)

        return statistics


    def calculate_sharpe_ratio(self, benchmark_return=0.00, N=252):
        """
        Calculate the sharpe ratio of our portfolio.
        Expects benchmark_return to be, for example, 0.01 for 1%
        """
        return_series = pd.Series(self.returns)
        return_annual = return_series.mean() * N
        sd_return_annual = return_series.std(ddof=1) * sqrt(N)
        return (return_annual - benchmark_return) / sd_return_annual


    def annualised_sharpe(self, returns, N=252):
        """
        Calculate the annualised Sharpe ratio of a returns stream
        based on a number of trading periods, N. N defaults to 252,
        which then assumes a stream of daily returns.
        The function assumes that the returns are the excess of
        those compared to a benchmark.
        """
        return np.sqrt(N) * returns.mean() / returns.std()


    def calculate_max_drawdown_pct(self):
        """
        Calculate the percentage drop related to the "worst"
        drawdown seen.
        """
        drawdown_series = pd.Series(self.drawdowns)
        equity_series = pd.Series(self.equity)
        bottom_index = drawdown_series.idxmax()
        try:
            top_index = equity_series[:bottom_index].idxmax()
            pct = (
                (equity_series.ix[top_index] - equity_series.ix[bottom_index]) /
                equity_series.ix[top_index] * 100
            )
            return round(pct, 4)
        except ValueError:
            return np.nan


    def plot_results(self):
        """
        A simple script to plot the balance of the portfolio, or
        "equity curve", as a function of time.
        """
        fig, axs = plt.subplots(2)
        axs[0].title.set_text('Equity Curve')
        axs[1].title.set_text('Drawdowns vs Time')
        axs[0].set(xlabel="Time",ylabel="Portfolio Value ($)")
        axs[1].set(xlabel="Time",ylabel="Drawdowns (%)")

        axs[0].plot(self.timeseries, self.asset_value)
        drawdownper = [ele*100 for ele in self.drawdowns]
        axs[1].plot(self.timeseries, drawdownper)
        plt.show()               




"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""


class Object(object):
    def __str__(self):
        return "Object"

    def __repr__(self):
        return str(id(self)) + ": " + self.__str__()



from data_handler.base import AbstractDataHandler
import queue
import time
import requests


class TwelveRealTimeBarHandler(AbstractDataHandler):
    """
    Get real time bar data using twelve data api
    https://rapidapi.com/twelvedata/api/twelve-data1/
    """
    def __init__(self, timeframe, event_queue, 
                 db_client=None
        ):
        self.db_client = db_client

                
    def get_next(self):
        pass
    
    
with open("/home/ron/Desktop/apikey", "r") as f:
    api_key = f.read()[:-1]
  

url = "https://twelve-data1.p.rapidapi.com/time_series"

querystring = {"symbol":"AMZN","interval":"1day","outputsize":"30","format":"json"}

headers = {
	"X-RapidAPI-Key": api_key,
	"X-RapidAPI-Host": "twelve-data1.p.rapidapi.com"
}

response = requests.request("GET", url, headers=headers, params=querystring)

print(response.text)
    
    
    


                                   
      


from abc import ABC, abstractmethod
from ..calendar.calendar import Calendlar
from datetime import date, timedelta


class BaseSchedule(ABC):
    def __init__(self):
        self.calendar = Calendlar()
 
    @abstractmethod
    def get_trading_hours(self, date):
        pass   
   
    def _gen_weekends(self, year):
        d = date(year, 1, 1)
        while d != date(year + 1, 1, 1):
            if 6 <= d.isoweekday() <= 7:
                self.calendar.add_holiday(d)
            d += timedelta(days=1)
            
    
            
      


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Nov 13 19:32:15 2021

@author: ron
"""

from .base import AbstractTechnicalIndicator
from collections import deque
import numpy as np

class SMA(AbstractTechnicalIndicator):
    
    def __init__(self, period):
        """
        Parameters
        ----------
        ready: to indicate whether data is sufficent to calucalte the result 
        """        
        self.data = deque(maxlen=period)
        self.data_store = []
        self.lastest_val = 0
        self.period = period
        self.ready = False

    def update(self, price):
        self.data.append(price)      
        #data is sufficient
        if not self.ready and len(self.data) == self.period:
            self.ready = True
            
        if self.ready:
            self.lastest_val = np.mean(self.data)
            self.data_store.append(self.lastest_val)

    
    def __bool__(self):
        return self.ready




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Nov 28 12:51:56 2021
@author: ron
"""
import unittest
from datetime import date, time
from tradingsystem.trading_schedule.us_stock_schedule import USStockSchedule
from tradingsystem.calendar.calendar import Calendlar


class MockUSStockSchedule(USStockSchedule):
    """
    since the parent __init__ method calls _config()
    which generates federal holidays and weekends,
    this mock class orverrides the __init__ method
    so that _gen_federal_holidays() and _gen_weekends()
    can be tested individually.
    """
    def __init__(self):
        self.calendar = Calendlar()
        self.close_early_date = set()
        self.summer_start_date = None
        self.winter_start_date = None
        

class TestUSStockSchedule(unittest.TestCase):
    """
    Test if the trading hours and public holidays are generated correctly for a given year.
    The result is compared with holidays listed in https://www.nyse.com/markets/hours-calendars
    """        
    def test_federal_holidays_2023(self):
        stock_schedule = MockUSStockSchedule()
        stock_schedule._gen_federal_holidays(2023)  
        calendar = stock_schedule.calendar
        
        new_year = date(2023, 1, 2)
        martin = date(2023, 1, 16)
        washington = date(2023, 2, 20)
        good_friday = date(2023, 4, 7)
        memorial_day = date(2023, 5, 29)
        juneteenth = date(2023, 6, 19)
        independence = date(2023, 7, 4)
        labor = date(2023, 9, 4)
        thanksgiving = date(2023, 11, 23)
        christams = date(2023, 12, 25)
        
        self.assertTrue(calendar.is_holiday(new_year))
        self.assertTrue(calendar.is_holiday(martin))
        self.assertTrue(calendar.is_holiday(washington))
        self.assertTrue(calendar.is_holiday(good_friday))
        self.assertTrue(calendar.is_holiday(memorial_day))
        self.assertTrue(calendar.is_holiday(juneteenth))
        self.assertTrue(calendar.is_holiday(independence))
        self.assertTrue(calendar.is_holiday(labor))
        self.assertTrue(calendar.is_holiday(thanksgiving))
        self.assertTrue(calendar.is_holiday(christams))

        
    def test_new_years_sat(self):
        """
        if new years is on sat, it should not precede to Friday
        """
        stock_schedule = MockUSStockSchedule()
        stock_schedule._gen_federal_holidays(2022)  
        calendar = stock_schedule.calendar
        new_year = date(2022, 1, 1)      
        self.assertFalse(calendar.is_holiday(new_year))

        
    def test_christmas_sat(self):
        """
        if christmas is on sat, it moves to 24th and the market
        should be closed on that day rather than closed early 
        """
        stock_schedule = MockUSStockSchedule()
        stock_schedule._gen_federal_holidays(2021)   
        calendar = stock_schedule.calendar
        christmas = date(2021, 12, 24)      
        self.assertTrue(calendar.is_holiday(christmas))
        self.assertFalse(christmas in stock_schedule.close_early_date)
        
    
    def test_close_early(self):
        """
        Test if some dates before/after the public holidays
        are regards as dates that the maket closes early
        """
        stock_schedule = MockUSStockSchedule()
        stock_schedule._gen_federal_holidays(2024)
        close_early = stock_schedule.close_early_date
        before_independence = date(2024, 7, 3)  
        after_thanksgiving = date(2024, 11, 29)  
        christams_Eve = date(2024, 12, 24)  
        self.assertTrue(before_independence in close_early)
        self.assertTrue(after_thanksgiving in close_early)
        self.assertTrue(christams_Eve in close_early)
        
    
    def test_daylight_saving_date(self):
        stock_schedule = MockUSStockSchedule()
        stock_schedule.summer_start_date = stock_schedule._find_nth_day(2022, 3, 6, 2)
        stock_schedule.winter_start_date = stock_schedule._find_nth_day(2022, 11, 6, 1)
        summer_time = date(2022, 3, 13)
        winter_time = date(2022, 11, 6)
        self.assertEqual(summer_time, stock_schedule.summer_start_date)
        self.assertEqual(winter_time, stock_schedule.winter_start_date)
    
    
    def test_gen_weekends(self):
        """
        test if all weekends generated correctly for a given year
        """
        stock_schedule = MockUSStockSchedule()
        stock_schedule._gen_weekends(2022)
        calendar = stock_schedule.calendar
        #there are 105 sundays and saturdays in 2022
        self.assertEqual(105, len(calendar.holidays))
        
    
    def test_trading_hours(self):
        stock_schedule = MockUSStockSchedule()
        stock_schedule._gen_federal_holidays(2023)
        stock_schedule._gen_weekends(2023)
        stock_schedule.summer_start_date = stock_schedule._find_nth_day(2023, 3, 6, 2)
        stock_schedule.winter_start_date = stock_schedule._find_nth_day(2023, 11, 6, 1)
                
        normal_summer_hours = stock_schedule.get_trading_hours(date(2023, 3, 15))
        early_summer_hours = stock_schedule.get_trading_hours(date(2023, 7, 3))
        normal_winter_hours = stock_schedule.get_trading_hours(date(2023, 1, 10))
        early_winter_hours = stock_schedule.get_trading_hours(date(2023, 11, 24))
        
        # self.assertEqual(normal_summer_hours, (time(13,30), time(20,00)))
        # self.assertEqual(early_summer_hours, (time(13,30), time(17,00))) 
        # self.assertEqual(normal_winter_hours, (time(14,30), time(21,00))) 
        # self.assertEqual(early_winter_hours, (time(14,30), time(18,00))) 
        self.assertEqual(normal_summer_hours, (1678887000, 1678910400))
        self.assertEqual(early_summer_hours, (1688391000, 1688403600))
        self.assertEqual(normal_winter_hours, (1673361000, 1673384400))
        self.assertEqual(early_winter_hours, (1700836200, 1700848800)) 
        
        

if __name__ == '__main__':
    unittest.main(verbosity=2)  


"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""

from . import order_condition
from .object_implem import Object
from .utils import * # @UnusedWildImport
from .server_versions import * # @UnusedWildImport
from .order import OrderComboLeg
from .contract import ComboLeg
from .tag_value import TagValue
from .wrapper import DeltaNeutralContract
from .softdollartier import SoftDollarTier

logger = logging.getLogger(__name__)


class OrderDecoder(Object):
    def __init__(self, contract, order, orderState, version, serverVersion):
        self.contract = contract
        self.order = order
        self.orderState = orderState
        self.version = version
        self.serverVersion = serverVersion

    def decodeOrderId(self, fields):
        self.order.orderId = decode(int, fields)

    def decodeContractFields(self, fields):
        self.contract.conId = decode(int, fields)
        self.contract.symbol = decode(str, fields)
        self.contract.secType = decode(str, fields)
        self.contract.lastTradeDateOrContractMonth = decode(str, fields)
        self.contract.strike = decode(float, fields)
        self.contract.right = decode(str, fields)
        if self.version >= 32:
            self.contract.multiplier = decode(str, fields)
        self.contract.exchange = decode(str, fields)
        self.contract.currency = decode(str, fields)
        self.contract.localSymbol = decode(str, fields)
        if self.version >= 32:
            self.contract.tradingClass = decode(str, fields)

    def decodeAction(self, fields):
        self.order.action = decode(str, fields)

    def decodeTotalQuantity(self, fields):
        self.order.totalQuantity = decode(Decimal, fields)

    def decodeOrderType(self, fields):
        self.order.orderType = decode(str, fields)

    def decodeLmtPrice(self, fields):
        if self.version < 29:
            self.order.lmtPrice = decode(float, fields)
        else:
            self.order.lmtPrice = decode(float, fields, SHOW_UNSET)

    def decodeAuxPrice(self, fields):
        if self.version < 30:
            self.order.auxPrice = decode(float, fields)
        else:
            self.order.auxPrice = decode(float, fields, SHOW_UNSET)

    def decodeTIF(self, fields):
        self.order.tif = decode(str, fields)
        
    def decodeOcaGroup(self, fields):
        self.order.ocaGroup = decode(str, fields)
        
    def decodeAccount(self, fields):
        self.order.account = decode(str, fields)

    def decodeOpenClose(self, fields):
        self.order.openClose = decode(str, fields)

    def decodeOrigin(self, fields):
        self.order.origin = decode(int, fields)

    def decodeOrderRef(self, fields):
        self.order.orderRef = decode(str, fields)
        
    def decodeClientId(self, fields):
        self.order.clientId = decode(int, fields)
        
    def decodePermId(self, fields):
        self.order.permId = decode(int, fields)

    def decodeOutsideRth(self, fields):
        self.order.outsideRth = decode(bool, fields)

    def decodeHidden(self, fields):
        self.order.hidden = decode(bool, fields)

    def decodeDiscretionaryAmt(self, fields):
        self.order.discretionaryAmt = decode(float, fields)

    def decodeGoodAfterTime(self, fields):
        self.order.goodAfterTime = decode(str, fields)
        
    def skipSharesAllocation(self, fields):
        _sharesAllocation = decode(str, fields) # deprecated

    def decodeFAParams(self, fields):
        self.order.faGroup = decode(str, fields)
        self.order.faMethod = decode(str, fields)
        self.order.faPercentage = decode(str, fields)
        self.order.faProfile = decode(str, fields)

    def decodeModelCode(self, fields):
        if self.serverVersion >= MIN_SERVER_VER_MODELS_SUPPORT:
            self.order.modelCode = decode(str, fields)

    def decodeGoodTillDate(self, fields):
        self.order.goodTillDate = decode(str, fields)

    def decodeRule80A(self, fields):
        self.order.rule80A = decode(str, fields)
        
    def decodePercentOffset(self, fields):
        self.order.percentOffset = decode(float, fields, SHOW_UNSET)
        
    def decodeSettlingFirm(self, fields):
        self.order.settlingFirm = decode(str, fields)

    def decodeShortSaleParams(self, fields):
        self.order.shortSaleSlot = decode(int, fields)
        self.order.designatedLocation = decode(str, fields)
        if self.serverVersion == MIN_SERVER_VER_SSHORTX_OLD:
            decode(int, fields)
        elif self.version >= 23:
            self.order.exemptCode = decode(int, fields)

    def decodeAuctionStrategy(self, fields):
        self.order.auctionStrategy = decode(int, fields)

    def decodeBoxOrderParams(self, fields):
        self.order.startingPrice = decode(float, fields, SHOW_UNSET)
        self.order.stockRefPrice = decode(float, fields, SHOW_UNSET)
        self.order.delta = decode(float, fields, SHOW_UNSET)

    def decodePegToStkOrVolOrderParams(self, fields):
        self.order.stockRangeLower = decode(float, fields, SHOW_UNSET)
        self.order.stockRangeUpper = decode(float, fields, SHOW_UNSET)

    def decodeDisplaySize(self, fields):
        self.order.displaySize = decode(int, fields, SHOW_UNSET)

    def decodeBlockOrder(self, fields):
        self.order.blockOrder = decode(bool, fields)

    def decodeSweepToFill(self, fields):
        self.order.sweepToFill = decode(bool, fields)

    def decodeAllOrNone(self, fields):
        self.order.allOrNone = decode(bool, fields)

    def decodeMinQty(self, fields):
        self.order.minQty = decode(int, fields, SHOW_UNSET)
        
    def decodeOcaType(self, fields):
        self.order.ocaType = decode(int, fields)

    def skipETradeOnly(self, fields):
        _eTradeOnly = decode(bool, fields) # deprecated

    def skipFirmQuoteOnly(self, fields):
        _firmQuoteOnly = decode(bool, fields) #` deprecated

    def skipNbboPriceCap(self, fields):
        _nbboPriceCap = decode(float, fields, SHOW_UNSET) # deprecated
        
    def decodeParentId(self, fields):
        self.order.parentId = decode(int, fields)
                
    def decodeTriggerMethod(self, fields):
        self.order.triggerMethod = decode(int, fields)


    def decodeVolOrderParams(self, fields, readOpenOrderAttribs):
        self.order.volatility = decode(float, fields, SHOW_UNSET)
        self.order.volatilityType = decode(int, fields)
        self.order.deltaNeutralOrderType = decode(str, fields)
        self.order.deltaNeutralAuxPrice = decode(float, fields, SHOW_UNSET)

        if self.version >= 27 and self.order.deltaNeutralOrderType:
            self.order.deltaNeutralConId = decode(int, fields)
            if readOpenOrderAttribs:
                self.order.deltaNeutralSettlingFirm = decode(str, fields)
                self.order.deltaNeutralClearingAccount = decode(str, fields)
                self.order.deltaNeutralClearingIntent = decode(str, fields)

        if self.version >= 31 and self.order.deltaNeutralOrderType:
            if readOpenOrderAttribs:
                self.order.deltaNeutralOpenClose = decode(str, fields)
            self.order.deltaNeutralShortSale = decode(bool, fields)
            self.order.deltaNeutralShortSaleSlot = decode(int, fields)
            self.order.deltaNeutralDesignatedLocation = decode(str, fields)

        self.order.continuousUpdate = decode(bool, fields)
        self.order.referencePriceType = decode(int, fields)

    def decodeTrailParams(self, fields):
        self.order.trailStopPrice = decode(float, fields, SHOW_UNSET)
        if self.version >= 30:
            self.order.trailingPercent = decode(float, fields, SHOW_UNSET)

    def decodeBasisPoints(self, fields):
        self.order.basisPoints = decode(float, fields, SHOW_UNSET)
        self.order.basisPointsType = decode(int, fields, SHOW_UNSET)
        
    def decodeComboLegs(self, fields):
        self.contract.comboLegsDescrip = decode(str, fields)

        if self.version >= 29:
            comboLegsCount = decode(int, fields)

            if comboLegsCount > 0:
                self.contract.comboLegs = []
                for _ in range(comboLegsCount):
                    comboLeg = ComboLeg()
                    comboLeg.conId = decode(int, fields)
                    comboLeg.ratio = decode(int, fields)
                    comboLeg.action = decode(str, fields)
                    comboLeg.exchange = decode(str, fields)
                    comboLeg.openClose = decode(int, fields)
                    comboLeg.shortSaleSlot = decode(int, fields)
                    comboLeg.designatedLocation = decode(str, fields)
                    comboLeg.exemptCode = decode(int, fields)
                    self.contract.comboLegs.append(comboLeg)

            orderComboLegsCount = decode(int, fields)
            if orderComboLegsCount > 0:
                self.order.orderComboLegs = []
                for _ in range(orderComboLegsCount):
                    orderComboLeg = OrderComboLeg()
                    orderComboLeg.price = decode(float, fields, SHOW_UNSET)
                    self.order.orderComboLegs.append(orderComboLeg)


    def decodeSmartComboRoutingParams(self, fields):
        if self.version >= 26:
            smartComboRoutingParamsCount = decode(int, fields)
            if smartComboRoutingParamsCount > 0:
                self.order.smartComboRoutingParams = []
                for _ in range(smartComboRoutingParamsCount):
                    tagValue = TagValue()
                    tagValue.tag = decode(str, fields)
                    tagValue.value = decode(str, fields)
                    self.order.smartComboRoutingParams.append(tagValue)

    def decodeScaleOrderParams(self, fields):
        if self.version >= 20:
            self.order.scaleInitLevelSize = decode(int, fields, SHOW_UNSET)
            self.order.scaleSubsLevelSize = decode(int, fields, SHOW_UNSET)
        else:
            self.order.notSuppScaleNumComponents = decode(int, fields, SHOW_UNSET)
            self.order.scaleInitLevelSize = decode(int, fields, SHOW_UNSET)

        self.order.scalePriceIncrement = decode(float, fields, SHOW_UNSET)

        if self.version >= 28 and self.order.scalePriceIncrement != UNSET_DOUBLE \
                and self.order.scalePriceIncrement > 0.0:
            self.order.scalePriceAdjustValue = decode(float, fields, SHOW_UNSET)
            self.order.scalePriceAdjustInterval = decode(int, fields, SHOW_UNSET)
            self.order.scaleProfitOffset = decode(float, fields, SHOW_UNSET)
            self.order.scaleAutoReset = decode(bool, fields)
            self.order.scaleInitPosition = decode(int, fields, SHOW_UNSET)
            self.order.scaleInitFillQty = decode(int, fields, SHOW_UNSET)
            self.order.scaleRandomPercent = decode(bool, fields)


    def decodeHedgeParams(self, fields):
        if self.version >= 24:
            self.order.hedgeType = decode(str, fields)
            if self.order.hedgeType:
                self.order.hedgeParam = decode(str, fields)

    def decodeOptOutSmartRouting(self, fields):
        if self.version >= 25:
            self.order.optOutSmartRouting = decode(bool, fields)

    def decodeClearingParams(self, fields):
        self.order.clearingAccount = decode(str, fields)
        self.order.clearingIntent = decode(str, fields)

    def decodeNotHeld(self, fields):
        if self.version >= 22:
            self.order.notHeld = decode(bool, fields)
            
    def decodeDeltaNeutral(self, fields):
        if self.version >= 20:
            deltaNeutralContractPresent = decode(bool, fields)
            if deltaNeutralContractPresent:
                self.contract.deltaNeutralContract = DeltaNeutralContract()
                self.contract.deltaNeutralContract.conId = decode(int, fields)
                self.contract.deltaNeutralContract.delta = decode(float, fields)
                self.contract.deltaNeutralContract.price = decode(float, fields)
            
    def decodeAlgoParams(self, fields):
        if self.version >= 21:
            self.order.algoStrategy = decode(str, fields)
            if self.order.algoStrategy:
                algoParamsCount = decode(int, fields)
                if algoParamsCount > 0:
                    self.order.algoParams = []
                    for _ in range(algoParamsCount):
                        tagValue = TagValue()
                        tagValue.tag = decode(str, fields)
                        tagValue.value = decode(str, fields)
                        self.order.algoParams.append(tagValue)

    def decodeSolicited(self, fields):
        if self.version >= 33:
            self.order.solicited = decode(bool, fields)

    def decodeOrderStatus(self, fields):
        self.orderState.status = decode(str, fields)

    def decodeWhatIfInfoAndCommission(self, fields):
        self.order.whatIf = decode(bool, fields)
        OrderDecoder.decodeOrderStatus(self, fields)
        if self.serverVersion >= MIN_SERVER_VER_WHAT_IF_EXT_FIELDS:
            self.orderState.initMarginBefore = decode(str, fields)
            self.orderState.maintMarginBefore = decode(str, fields)
            self.orderState.equityWithLoanBefore = decode(str, fields)
            self.orderState.initMarginChange = decode(str, fields)
            self.orderState.maintMarginChange = decode(str, fields)
            self.orderState.equityWithLoanChange = decode(str, fields)

        self.orderState.initMarginAfter = decode(str, fields)
        self.orderState.maintMarginAfter = decode(str, fields)
        self.orderState.equityWithLoanAfter = decode(str, fields)

        self.orderState.commission = decode(float, fields, SHOW_UNSET)
        self.orderState.minCommission = decode(float, fields, SHOW_UNSET)
        self.orderState.maxCommission = decode(float, fields, SHOW_UNSET)
        self.orderState.commissionCurrency = decode(str, fields)
        self.orderState.warningText = decode(str, fields)

    def decodeVolRandomizeFlags(self, fields):
        if self.version >= 34:
            self.order.randomizeSize = decode(bool, fields)
            self.order.randomizePrice = decode(bool, fields)
        
    def decodePegToBenchParams(self, fields):
        if self.serverVersion >= MIN_SERVER_VER_PEGGED_TO_BENCHMARK:
            if self.order.orderType == "PEG BENCH":
                self.order.referenceContractId = decode(int, fields)
                self.order.isPeggedChangeAmountDecrease = decode(bool, fields)
                self.order.peggedChangeAmount = decode(float, fields)
                self.order.referenceChangeAmount = decode(float, fields)
                self.order.referenceExchangeId = decode(str, fields)
        
    def decodeConditions(self, fields):
        if self.serverVersion >= MIN_SERVER_VER_PEGGED_TO_BENCHMARK:
            conditionsSize = decode(int, fields)
            if conditionsSize > 0:
                self.order.conditions = []
                for _ in range(conditionsSize):
                    conditionType = decode(int, fields)
                    condition = order_condition.Create(conditionType)
                    condition.decode(fields)
                    self.order.conditions.append(condition)

                self.order.conditionsIgnoreRth = decode(bool, fields)
                self.order.conditionsCancelOrder = decode(bool, fields)
        
            
    def decodeAdjustedOrderParams(self, fields):
        if self.serverVersion >= MIN_SERVER_VER_PEGGED_TO_BENCHMARK:
            self.order.adjustedOrderType = decode(str, fields)
            self.order.triggerPrice = decode(float, fields)
            OrderDecoder.decodeStopPriceAndLmtPriceOffset(self, fields)
            self.order.adjustedStopPrice = decode(float, fields)
            self.order.adjustedStopLimitPrice = decode(float, fields)
            self.order.adjustedTrailingAmount = decode(float, fields)
            self.order.adjustableTrailingUnit = decode(int, fields)
            
    def decodeStopPriceAndLmtPriceOffset(self, fields):
            self.order.trailStopPrice = decode(float, fields)
            self.order.lmtPriceOffset = decode(float, fields)
            
    def decodeSoftDollarTier(self, fields):
        if self.serverVersion >= MIN_SERVER_VER_SOFT_DOLLAR_TIER:
            name = decode(str, fields)
            value = decode(str, fields)
            displayName = decode(str, fields)
            self.order.softDollarTier = SoftDollarTier(name, value, displayName)

    def decodeCashQty(self, fields):
        if self.serverVersion >= MIN_SERVER_VER_CASH_QTY:
            self.order.cashQty = decode(float,fields)

    def decodeDontUseAutoPriceForHedge(self, fields):
        if self.serverVersion >= MIN_SERVER_VER_AUTO_PRICE_FOR_HEDGE:
            self.order.dontUseAutoPriceForHedge = decode(bool,fields)

    def decodeIsOmsContainers(self, fields):
        if self.serverVersion >= MIN_SERVER_VER_ORDER_CONTAINER:
            self.order.isOmsContainer = decode(bool, fields)

    def decodeDiscretionaryUpToLimitPrice(self, fields):
        if self.serverVersion >= MIN_SERVER_VER_D_PEG_ORDERS:
            self.order.discretionaryUpToLimitPrice = decode(bool, fields)

    def decodeAutoCancelDate(self, fields):
        self.order.autoCancelDate = decode(str, fields)

    def decodeFilledQuantity(self, fields):
        self.order.filledQuantity = decode(Decimal, fields)

    def decodeRefFuturesConId(self, fields):
        self.order.refFuturesConId = decode(int, fields)

    def decodeAutoCancelParent(self, fields, minVersionAutoCancelParent = MIN_CLIENT_VER):
        if self.serverVersion >= minVersionAutoCancelParent:
            self.order.autoCancelParent = decode(bool, fields)

    def decodeShareholder(self, fields):
        self.order.shareholder = decode(str, fields)

    def decodeImbalanceOnly(self, fields):
        self.order.imbalanceOnly = decode(bool, fields)

    def decodeRouteMarketableToBbo(self, fields):
        self.order.routeMarketableToBbo = decode(bool, fields)

    def decodeParentPermId(self, fields):
        self.order.parentPermId = decode(int, fields)

    def decodeCompletedTime(self, fields):
        self.orderState.completedTime = decode(str, fields)

    def decodeCompletedStatus(self, fields):
        self.orderState.completedStatus = decode(str, fields)

    def decodeUsePriceMgmtAlgo(self, fields):
        if self.serverVersion >= MIN_SERVER_VER_PRICE_MGMT_ALGO:
            self.order.usePriceMgmtAlgo = decode(bool, fields)
            
    def decodeDuration(self, fields):
        if self.serverVersion >= MIN_SERVER_VER_DURATION:
            self.order.duration = decode(int, fields, SHOW_UNSET)
            
    def decodePostToAts(self, fields):
        if self.serverVersion >= MIN_SERVER_VER_POST_TO_ATS:
            self.order.postToAts = decode(int, fields, SHOW_UNSET)



from enum import Enum

EventType = Enum("EventType", "BAR SIGNAL ORDER FILL")
OrderType = Enum("OrderType", "MARKET LIMIT") 
Direction = Enum("Direction", "LONG SHORT")


class event:
    """
    event is base class providing an interface for all subsequent
    (inherited) events, that will trigger further events in the
    trading infrastructure.
    """
    @property
    def get_event_type(self):
        return self.type.name


class BarEvent(event):
    """
    Created by a DataHandler object.
    This is received by a Strategy object and Portfolio and acted upon.
    """
    def __init__(
        self, timestamp, ticker,
        open_price, high_price, low_price,
        close_price, volume
        ):
        """
        Initialises the BarEvent.
        Parameters:
        timestamp - The timestamp of the Bar event generated (i.e. timestamp of a bar)
        ticker - The ticker symbol, e.g. 'GOOG'.
        open_price - The opening price of the bar
        high_price - The high price of the bar
        low_price - The low price of the bar
        close_price - The close price of the bar
        volume - The volume of trading within the bar
        """
        self.type = EventType.BAR
        self.timestamp = timestamp
        self.ticker = ticker
        self.open_price = open_price
        self.high_price = high_price
        self.low_price = low_price
        self.close_price = close_price
        self.volume = volume

    def __str__(self):
        format_str = "Type: %s, Time: %s, Ticker: %s, " \
            "Open: %s, High: %s, Low: %s, Close: %s, Volume: %s" % (
                str(self.type), str(self.timestamp), str(self.ticker),
                str(self.open_price), str(self.high_price), 
                str(self.low_price), str(self.close_price),
                str(self.volume)
            )
        return format_str

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return self.type == other.type and self.timestamp == other.timestamp and \
            self.ticker == other.ticker and self.open_price == other.open_price and \
                self.high_price == other.high_price and self.low_price == other.low_price and \
                    self.close_price == other.close_price and self.volume == other.volume



class SignalEvent(event):
    """
    Created by a Strategy object.
    This is received by a Portfolio object and acted upon.
    """
    def __init__(self, timestamp, ticker, order_type, 
                 direction, bar_closing_price, entry_price=None
                 ):
        """
        Initialises the SignalEvent.
        Parameters:
        timestamp - timestamp of trading signal generated
        ticker - The ticker symbol, e.g. 'GOOG'.
        order_type - Market order or Limit order
        direction - 'LONG', 'SHORT', 'LONG_EXIT' or 'SHORT_EXIT'
        entry_price - entry price of the ticket, it should be None if 
                      it is a market order
        bar_closing_price - used by Position Handler to determine the number of
                            shares should be bought if it is market order.
        """
        self.type = EventType.SIGNAL
        self.timestamp = timestamp
        self.ticker = ticker
        self.order_type = order_type
        self.direction = direction
        self.bar_closing_price = bar_closing_price
        self.entry_price = entry_price
        #self.entry_price_range = entry_price_range
        
        
    def __str__(self):
        format_str = "Type: %s, Time: %s, Ticker: %s, Order Type: %s, " \
            "Direction: %s, Entry price: %s, Closing price: %s" % (
                str(self.type), str(self.timestamp), str(self.ticker),
                str(self.order_type), str(self.direction), 
                str(self.entry_price), str(self.bar_closing_price)
            )
        return format_str

    def __repr__(self):
        return str(self)
        


class OrderEvent(event):
    """
    Created by a Portfolio object.
    This is received by a Broker object and acted upon.
    """
    def __init__(self, timestamp, ticker, order_type, 
                 direction, quantity, entry_price=None
                 ):
        """
        Initialises the OrderEvent.
        Parameters:
        timestamp - timestamp of order event generated
        ticker - The ticker symbol, e.g. 'GOOG'.
        direction - 'LONG', 'SHORT', 'LONG_EXIT' or 'SHORT_EXIT'
        order_type - Market order or Limit order
        entry_price - entry price of the ticket, it should be None if 
                      it is a market order
        quantity - The quantity of shares to transact.
        """
        self.type = EventType.ORDER
        self.timestamp = timestamp
        self.ticker = ticker
        self.order_type = order_type
        self.direction = direction
        self.quantity = quantity
        self.entry_price = entry_price
        #self.entry_price_range = entry_price_range

    def __str__(self):
        format_str = "Type: %s, Time: %s, Ticker: %s, Order Type: %s, " \
            "Direction: %s, Entry price: %s, Quantity: %s" % (
                str(self.type), str(self.timestamp), str(self.ticker),
                str(self.order_type), str(self.direction), 
                str(self.entry_price), str(self.quantity)
            )
        return format_str

    def __repr__(self):
        return str(self)


class FillEvent(event):
    """
    Created by a Broker object.
    This is received by a Portfolio object to update the portfolio
    """

    def __init__(
        self, timestamp, ticker,
        direction, exchange,
        price, quantity,
        commission
        ):
        """
        Initialises the FillEvent object.
        timestamp - The timestamp when the order was filled.
        ticker - The ticker symbol, e.g. 'GOOG'.
        direction - 'LONG', 'SHORT', 'LONG_EXIT' or 'SHORT_EXIT'
        quantity - The filled quantity.
        exchange - The exchange where the order was filled.
        price - The price at which the trade was filled
        commission - The brokerage commission for carrying out the trade.
        """
        self.type = EventType.FILL
        self.timestamp = timestamp
        self.ticker = ticker
        self.direction = direction
        self.exchange = exchange
        self.price = price
        self.quantity = quantity
        self.commission = commission
        
    def __str__(self):
        format_str = "Type: %s, Time: %s, Ticker: %s, " \
            "Direction: %s, Exchange: %s, Price: %s, " \
            "Quantity: %s, Commission: %s" % (
                str(self.type), str(self.timestamp), str(self.ticker),
                str(self.direction), str(self.exchange), 
                str(self.price), str(self.quantity), str(self.commission)
            )
        return format_str

    def __repr__(self):
        return str(self)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov 22 20:12:56 2021

@author: ron
"""
import unittest
from tradingsystem.event.event import BarEvent, SignalEvent, OrderEvent, FillEvent, OrderType, EventType, Direction
from datetime import datetime


class TestEvent(unittest.TestCase):
    
    def setUp(self):
        self.date_time = datetime(2021, 11, 5, 15, 30, 20)

              
    def test_BarEvent(self):
         bar = BarEvent(self.date_time, 'GOOG', 2779.970, 2843.540, 2774.959, 2821.990, 105448)
         self.assertEqual(bar.type, EventType.BAR)
         self.assertEqual(bar.timestamp, self.date_time)
         self.assertEqual(bar.ticker, 'GOOG')
         self.assertEqual(bar.open_price, 2779.970)
         self.assertEqual(bar.high_price, 2843.540)
         self.assertEqual(bar.low_price, 2774.959)
         self.assertEqual(bar.close_price, 2821.990)
         self.assertEqual(bar.volume, 105448)
        

                
    def test_SignalEvent_Market(self):
         signal = SignalEvent(self.date_time, 'GOOG', OrderType.MARKET, Direction.LONG, 2821.990)
         self.assertEqual(signal.type, EventType.SIGNAL)
         self.assertEqual(signal.timestamp, self.date_time)
         self.assertEqual(signal.ticker, 'GOOG')
         self.assertEqual(signal.order_type, OrderType.MARKET)
         self.assertEqual(signal.direction, Direction.LONG)
         self.assertEqual(signal.bar_closing_price, 2821.990)
         self.assertEqual(signal.entry_price, None)
         
         
    def test_SignalEvent_Limit(self):
         signal = SignalEvent(self.date_time, 'GOOG', OrderType.LIMIT, Direction.LONG, 2821.990, 2820)
         self.assertEqual(signal.type, EventType.SIGNAL)
         self.assertEqual(signal.timestamp, self.date_time)
         self.assertEqual(signal.ticker, 'GOOG')
         self.assertEqual(signal.order_type, OrderType.LIMIT)
         self.assertEqual(signal.direction, Direction.LONG)
         self.assertEqual(signal.bar_closing_price, 2821.990)
         self.assertEqual(signal.entry_price, 2820)

        
    
    def test_OrderEvent(self):
         order = OrderEvent(self.date_time, 'GOOG', OrderType.MARKET, Direction.LONG, 100)
         self.assertEqual(order.type, EventType.ORDER)
         self.assertEqual(order.timestamp, self.date_time)
         self.assertEqual(order.ticker, 'GOOG')
         self.assertEqual(order.order_type, OrderType.MARKET)
         self.assertEqual(order.direction, Direction.LONG)
         self.assertEqual(order.quantity, 100)



    def test_FillEvent(self):   
         fill = FillEvent(self.date_time, 'GOOG', Direction.LONG, 'NASDAQ', 2821.990, 100, 1)
         self.assertEqual(fill.type, EventType.FILL)
         self.assertEqual(fill.timestamp, self.date_time)
         self.assertEqual(fill.ticker, 'GOOG')
         self.assertEqual(fill.exchange, 'NASDAQ')
         self.assertEqual(fill.price, 2821.990)
         self.assertEqual(fill.quantity, 100)
         self.assertEqual(fill.commission, 1)
        
        

if __name__ == '__main__':
    unittest.main(verbosity=2)




from .base import AbstractBroker
from ..event.event import FillEvent, OrderType

class IBSimulatedBroker(AbstractBroker):
    """
    The simulated execution handler for Interactive Brokers
    converts all order objects into their equivalent fill
    objects automatically without latency, slippage or
    fill-ratio issues.
    """
    def __init__(self, event_queue, data_handler, db_client=None):
        self.db_client = db_client
        self.event_queue = event_queue
        self.unfilled_order = {} #store limit order that cannot yet match the market price 
        self.data_handler = data_handler
 
    
    def execute_order(self, order_event):
        if order_event.order_type == OrderType.MARKET:
            timestamp = order_event.timestamp
            ticker = order_event.ticker
            direction = order_event.direction
            exchange = "Test Exchange"
            price = self.data_handler.get_last(ticker)
            quantity = order_event.quantity
            commission = 1
            fill_event = FillEvent(timestamp, ticker, direction,
                                   exchange, price, quantity, commission)
            self.event_queue.put(fill_event)
            print(fill_event)

                         
        elif order_event.order_type == OrderType.LIMIT:
            pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Dec  7 20:05:06 2021

@author: ron
"""


from ..ibapi.wrapper import EWrapper
import datetime
from ..event.event import Direction
from copy import deepcopy
from ..common import Bar, Error
from ..ibapi.common import TickerId, TickAttrib
from decimal import Decimal
from ..ibapi.ticktype import * # @UnusedWildImport
import time


class TWSWrapper(EWrapper):
    
    def __init__(self):
        EWrapper.__init__(self)


    def historicalData(self, reqId, bar):
        """ returns the requested historical data bars
        reqId - the request's identifier
        date  - the bar's date and time (either as a yyyymmss hh:mm:ssformatted
             string or as system time according to the request)
        open  - the bar's open point
        high  - the bar's high point
        low   - the bar's low point
        close - the bar's closing point
        volume - the bar's traded volume if available
        count - the number of trades during the bar's timespan (only available
            for TRADES).
        WAP -   the bar's Weighted Average Price
        hasGaps  -indicates if the data has gaps or not. """
        #--------------------------------------------------       
        bar.date += " GMT+0000"
        timestamp = int(datetime.datetime.strptime(bar.date, "%Y%m%d  %H:%M:%S GMT%z").timestamp()) #bar start time
 
        if timestamp >= self.last_bar_time:
            current_bar = Bar(timestamp, timestamp + 5, bar.open, bar.high,
                             bar.low, bar.close, bar.volume, bar.wap, bar.barCount)
            self._construct_bar(reqId, current_bar)

      
    def historicalDataEnd(self, reqId, start, end): 
        """ Marks the ending of the historical bars reception. """
        print('Start Date: {start_date}, End Date: {end_date}'.format(start_date=start, end_date=end))              
        with self.cond:
            self.cond.notify_all()

        
    def realtimeBar(self, reqId, time, open_, high, low, close, volume, wap, count):
        current_bar = Bar(time, time + 5, open_, high,
                          low, close, volume, wap, count)      

        if current_bar.end_time >= self.last_bar_time:
            self._construct_bar(reqId, current_bar) 

        if not self.realtime_subscribed:
            with self.cond:
                self.realtime_subscribed = True
                self.cond.notify_all()       


    def historicalDataUpdate(self, reqId, bar):
        #print("In historicalDataUpdate")
        """returns updates in real time when keepUpToDate is set to True"""
        bar.date += " GMT+0000"
        timestamp = int(datetime.datetime.strptime(bar.date, "%Y%m%d  %H:%M:%S GMT%z").timestamp())
        bar.date = timestamp
        print(reqId, bar)           
        current_bar = Bar(timestamp, timestamp + 5, bar.open, bar.high,
                           bar.low, bar.close, bar.volume, bar.wap, bar.barCount)


    def nextValidId(self, orderId):
        with self.cond:
            self.nextValidOrderId = orderId
            self.cond.notify_all()


    def openOrder(self, orderId, contract, order, orderState):
        super().openOrder(orderId, contract, order, orderState)
        #print("-------------------------------")
        # print("OpenOrder. PermId: ", order.permId, "ClientId:", order.clientId, " OrderId:", orderId, 
        #         "Account:", order.account, "Symbol:", contract.symbol, "SecType:", contract.secType,
        #         "Exchange:", contract.exchange, "Action:", order.action, "OrderType:", order.orderType,
        #         "TotalQty:", decimalMaxString(order.totalQuantity), "CashQty:", order.cashQty, 
        #         "LmtPrice:", order.lmtPrice, "AuxPrice:", order.auxPrice, "Status:", orderState.status) 

        
    def orderStatus(self, orderId, status, filled,
                    remaining, avgFillPrice, permId,
                    parentId, lastFillPrice, clientId,
                    whyHeld, mktCapPrice):
        super().orderStatus(orderId, status, filled,
                    remaining, avgFillPrice, permId,
                    parentId, lastFillPrice, clientId,
                    whyHeld, mktCapPrice)
        #print("-------------------------------")
        # print("OrderStatus. Id:", orderId, "Status:", status, "Filled:", decimalMaxString(filled),
        #         "Remaining:", decimalMaxString(remaining), "AvgFillPrice:", avgFillPrice,
        #         "PermId:", permId, "ParentId:", parentId, "LastFillPrice:",
        #         lastFillPrice, "ClientId:", clientId, "WhyHeld:",
        #         whyHeld, "MktCapPrice:", mktCapPrice)

        
    def execDetails(self, reqId, contract, execution):
        execution.time += " GMT+0000"
        timestamp = int(datetime.datetime.strptime(execution.time , "%Y%m%d  %H:%M:%S GMT%z").timestamp())        
        self.fill_event.timestamp = timestamp
        
        if contract.secType == "STK":       
            self.fill_event.ticker = contract.symbol
        elif contract.secType == "CASH":       
            self.fill_event.ticker = contract.symbol + "/" + contract.currency
            
        if execution.side == "BOT":
            self.fill_event.direction = Direction.LONG
        else:
            self.fill_event.direction = Direction.SHORT
            
        self.fill_event.exchange = execution.exchange
        self.fill_event.price = execution.price
        self.fill_event.quantity = int(execution.shares)

            
    def commissionReport(self, commissionReport):
        self.fill_event.commission = commissionReport.commission
        self.execution_detail.put(deepcopy(self.fill_event))
        #print("CommissionReport.", commissionReport)
        
 
    def completedOrder(self, contract, order, orderState):
        print("------------------------------------------")
        print("Contract:", contract)
        print("order:", order)
        """This function is called to feed in completed orders.

        contract: Contract - The Contract class attributes describe the contract.
        order: Order - The Order class gives the details of the completed order.
        orderState: OrderState - The orderState class includes completed order status details."""
   
    
    def headTimestamp(self, reqId:int, headTimestamp:str):
        """returns earliest available data of a type of data for a particular contract"""
        print(headTimestamp)
        
        
    def error(self, reqId, errorCode, errorString):
        """This event is called when there is an error with the
        communication or when TWS wants to send a message to the client."""
        super().error(reqId, errorCode, errorString)
        time_now  = time.time()
        error = Error(time_now, errorCode, errorString)
        self.error_code.put(errorCode)


    def currentTime(self, time: int):
        with self.cond:
            self.time = time
            self.cond.notify_all()
        # print("tws time:", time)


    def tickPrice(self, reqId:TickerId , tickType:TickType, price:float,
                  attrib:TickAttrib):
        """Market data tick price callback. Handles all price related ticks."""
        super().tickPrice(reqId, tickType, price, attrib)
        print("TickPrice. TickerId:", reqId, "tickType:", tickType,
            "Price:", str(price), "CanAutoExecute:", attrib.canAutoExecute,
            "PastLimit:", attrib.pastLimit, end=' ')
        if tickType == TickTypeEnum.BID or tickType == TickTypeEnum.ASK:
            print("PreOpen:", attrib.preOpen)
        else:
            print() 


    def tickSize(self, reqId:TickerId, tickType:TickType, size:Decimal):
        """Market data tick size callback. Handles all size-related ticks."""
        super().tickSize(reqId, tickType, size)
        #print("TickSize. TickerId:", reqId, "TickType:", tickType, "Size: ", str(size))


    def tickSnapshotEnd(self, reqId:int):
        """When requesting market data snapshots, this market will indicate the
        snapshot reception is finished. """
        super().tickSnapshotEnd(reqId)


    def tickGeneric(self, reqId: TickerId, tickType: TickType, value: float):
        super().tickGeneric(reqId, tickType, value)
        print("TickGeneric. TickerId:", reqId, "TickType:", tickType, "Value:", str(value))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Nov 20 14:35:19 2021

@author: ron
"""

from .base import AbstractTechnicalIndicator


class EMA(AbstractTechnicalIndicator):
    
    def __init__(self, period):
        self.period = period
        self.prev_ema = 0
        self.lastest_val = 0
        self.data_store = []
        self.ready = False
        self.multiplier = 2 / (period + 1)
        self.first_ema_count = 0


    def update(self, price):
        #data needed for the first EMA
        if self.first_ema_count < self.period:
            self.prev_ema += price
            self.first_ema_count += 1
            if self.first_ema_count == self.period:
                self.ready = True
            
        if self.ready:   
            if self.first_ema_count == self.period:
                self.lastest_val = self.prev_ema / self.period
                self.first_ema_count += 1
            else:
                self.lastest_val = (price - self.prev_ema) * self.multiplier + self.prev_ema
                
            self.prev_ema = self.lastest_val
            self.data_store.append(self.lastest_val)

                          
    def __bool__(self):
        return self.ready
            
            
                   
                    
            
                
                
                    
            
            
            
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Oct 29 18:22:08 2021

@author: ron
"""
import queue
from .event.event import EventType


class TradingSession:
    """
    Enscapsulates the settings and components for
    carrying out either a backtest or live trading session.
    """
    def __init__(
        self, price_handler, strategy, portfolio, 
        order_handler, broker, events_queue,
        statistics
        ):
        """
        Set up the backtest variables according to
        what has been passed in.
        """
        self.strategy = strategy
        self.events_queue = events_queue
        self.data_handler = price_handler
        self.portfolio = portfolio
        self.order_handler = order_handler
        self.broker = broker
        self.statistics = statistics
        self.previous_day = None


    def run_session(self):
        """
        Carries out an infinite while loop that polls the
        events queue and directs each event to either the
        strategy component of the execution handler. The
        loop continue until the event queue has been
        emptied.
        """
        while True:
            try:
                event = self.events_queue.get(False)
            except queue.Empty:
                try:
                    self.data_handler.get_next()
                except StopIteration:
                    #get the update of last day and break out the while loop
                    self.statistics.update(self.previous_day, self.portfolio.get_asset_val())                                  
                    break
            else:
                if self.previous_day is None:
                    self.previous_day = event.timestamp.date()
                elif event.timestamp.date() > self.previous_day:
                    self.statistics.update(self.previous_day, self.portfolio.get_asset_val())
                    self.previous_day = event.timestamp.date()  

                if event.type == EventType.BAR:
                    self.strategy.calculate_signal(event)
                    self.portfolio.on_bar(event)                  
                elif event.type == EventType.SIGNAL:
                    self.order_handler.create_order(event)
                elif event.type == EventType.ORDER:
                    self.broker.execute_order(event)
                elif event.type == EventType.FILL:
                    self.portfolio.on_fill(event)
                else:
                    raise NotImplementedError("Unsupported event.type '%s'" % event.type)


    def start_trading(self, testing=False):
        """
        Runs either a backtest or live session, and outputs performance when complete.
        """
        self.run_session()
        print("---------------------------------")
        print('Ending cash: ' + str(self.portfolio.cash))
        print('Ending market value: ' + str(self.portfolio.market_value))
        print('Ending asset value: ' + str(self.portfolio.get_asset_val()))
        results = self.statistics.get_results()
        print("---------------------------------")
        print("Strategy:", self.strategy)
        print("Return: %0.2f%%" % results["return"])
        print("Sharpe Ratio: %0.2f" % results["sharpe_ratio"])
        print(
            "Max Drawdown: %0.2f%%" % (
                results["max_drawdown"] * 100.0
            )
        )
        return_over_mdd = results["return"] / (results["max_drawdown"] * 100 * -1)
        print("Return Over Maximum Drawdown: %0.2f" % return_over_mdd)
        print("Backtest complete.")
        self.statistics.plot_results()
 







from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="tradingsystem",
    version="0.0.1",
    description="An event-driven trading system",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ronshen0404/tradingsystem",
    author="ronshen0404",
    author_email="ronshen970404@gmail.com",
    classifiers=[
        "Programming Language :: Python :: 3.9",
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux',
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "mysql-connector-python>=8.0.30",
        "matplotlib>=3.5.2",
        "numpy>=1.23.1",
        "pandas>=1.4.3",
        "yfinance>=0.1.74",
    ]
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Nov 17 20:41:09 2021

@author: ron
"""

from .base import AbstractStrategy
from ..technical_indicators.bb import BB
from ..technical_indicators.rsi import RSI
from ..event.event import SignalEvent, OrderType, Direction
from ..common import get_cur_time



class BBRSI(AbstractStrategy):
    
    def __init__(self, bb_period, multiple, rsi_period,
                 event_queue, session_type, 
                 long_only=True, stoploss=None, 
                 portfolio_handler=None,
                 db_client=None
        ):
        """
        Parameters
        ----------
        tickers_data : dict: {key:string, value: list}
                       key is the ticker symbol, e.g. 'GOOG'
                       list[0] stores BB object.
                       list[1] stores RSI object.
                       
        session_type: 'backtest' or 'live'
        -------

        """
        self.db_client = db_client
        self.bb_period = bb_period
        self.multiple = multiple
        self.rsi_period = rsi_period
        self.tickers_data = {}
        self.event_queue = event_queue
        self.invested = False
        self.session_type = session_type
        
        if stoploss is not None and portfolio_handler is None:
            raise ValueError("portfolio_handler must be set in order to calculate stop loss")
        else:
            self.stoploss = stoploss
            self.portfolio_handler = portfolio_handler

    
    def calculate_signal(self, bar_event):
        if not bar_event.ticker in self.tickers_data:
            self._subscribe_ticker(bar_event.ticker)
            
        self.tickers_data[bar_event.ticker][0].update(bar_event.close_price)
        self.tickers_data[bar_event.ticker][1].update(bar_event.close_price)
        
        if self.tickers_data[bar_event.ticker][0] and \
           self.tickers_data[bar_event.ticker][1]:
                             
            upper, mid, lower = self.tickers_data[bar_event.ticker][0].get_lastest_val()
            rsi = self.tickers_data[bar_event.ticker][1].get_lastest_val()
            
            if bar_event.close_price <= lower and rsi <= 30 and not self.invested:
                cur_time = get_cur_time(bar_event, self.session_type)
                signal_event = SignalEvent(cur_time, bar_event.ticker,
                                           OrderType.MARKET, Direction.LONG, 
                                           bar_event.close_price)
                                  
                self.event_queue.put(signal_event)
                self.invested = True
                print(signal_event)
                
            elif ((bar_event.close_price >= upper and rsi >= 70) or self._stop_loss_trigger(bar_event)) \
                 and self.invested:
                cur_time = get_cur_time(bar_event, self.session_type)
                signal_event = SignalEvent(cur_time, bar_event.ticker,
                                           OrderType.MARKET, Direction.EXIT, 
                                           bar_event.close_price)
                                    
                self.event_queue.put(signal_event)
                self.invested = False    
                print(signal_event)
             
    
    def _stop_loss_trigger(self, bar_event):
        if self.stoploss is None:
            return False
        else:
            avg_price = self.portfolio_handler.get_fin_instrucements(bar_event.ticker)["AVG_PRICE"]
            loss = (avg_price - bar_event.close_price) / avg_price
            if loss > self.stoploss:
                return True
            else:
                return False

            
    def _subscribe_ticker(self, ticker_name):
            bb = BB(self.bb_period, self.multiple)
            rsi = RSI(self.rsi_period)
            data = [bb, rsi]
            self.tickers_data[ticker_name] = data
            a = self.tickers_data[ticker_name]

        
    def __str__(self):
        return "BB-RSI Strategy"


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Nov 13 19:51:56 2021

@author: ron
"""

from .base import AbstractStrategy
from ..event.event import SignalEvent, OrderType, Direction
from ..technical_indicators.rsi import RSI
from ..common import get_cur_time


class RelativeStrengthIndex(AbstractStrategy):
    def __init__(self, ticker_list, period, 
                 event_queue, session_type, portfolio,
                 db_client=None
        ):
        """
        Parameters
        ----------
        ticker_list: The ticker symbols used for this strategy
        session_type: SessionType.LIVE or SessionType.BACKTEST
        ---------
        """
        """
        tickers_data: dict: {key:string, value: list}
                key is the ticker symbol, e.g. 'GOOG'
                list[0] stores short sma object.
                list[1] stores long sma object.
        """
        self.db_client = db_client
        self.period = period
        self.tickers_data = self._subscribe_ticker(ticker_list)
        self.event_queue = event_queue
        self.session_type = session_type
        self.portfolio = portfolio


    def calculate_signal(self, bar_event):        
        ticker_data = self.tickers_data[bar_event.ticker]  
        ticker_data.update(bar_event.close_price)
        
        if ticker_data:              
            rsi = ticker_data.lastest_val
            ticker_info = self.portfolio.get_fin_instrument(bar_event.ticker)
            #the ticker symbol has not been long/short yet
            if ticker_info is None:
                #go long
                if rsi <= 30:                    
                    cur_time = get_cur_time(bar_event, self.session_type)
                    signal_event = SignalEvent(cur_time, bar_event.ticker,
                                               OrderType.MARKET, Direction.LONG, 
                                               bar_event.close_price)
                    print(signal_event)
                    self.event_queue.put(signal_event)

                #go short   
                elif rsi >= 70:
                    cur_time = get_cur_time(bar_event, self.session_type)
                    signal_event = SignalEvent(cur_time, bar_event.ticker,
                                               OrderType.MARKET, Direction.SHORT, 
                                               bar_event.close_price)
                    print(signal_event)
                    self.event_queue.put(signal_event)

            #the ticker symbol has been long/short                                       
            else:
                #close long position 
                if ticker_info.POS > 0 and rsi >= 70:
                    cur_time = get_cur_time(bar_event, self.session_type)
                    signal_event = SignalEvent(cur_time, bar_event.ticker,
                                               OrderType.MARKET, Direction.SHORT, 
                                               bar_event.close_price)                                       
                    self.event_queue.put(signal_event)
                    print(signal_event)
                            
                #close short position            
                elif ticker_info.POS < 0 and rsi <= 30:                              
                    cur_time = get_cur_time(bar_event, self.session_type)
                    signal_event = SignalEvent(cur_time, bar_event.ticker,
                                            OrderType.MARKET, Direction.LONG, 
                                            bar_event.close_price)                
                    self.event_queue.put(signal_event)
                    print(signal_event)              

                        
    def _subscribe_ticker(self, ticker_list):
        tickers_data = {}
        for ticker in ticker_list:
            tickers_data[ticker] = RSI(self.period) 
        return tickers_data                                                                                        

            
    def get_ticker_data(self, ticker):
        if not ticker in self.tickers_data:
            return None

        return self.tickers_data[ticker]

    
    def __str__(self):
        return "RSI"
    
        
            



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Nov 27 15:27:07 2021

@author: ron
"""

import unittest
from tradingsystem.event.event import BarEvent, FillEvent, Direction
from tradingsystem.portfolio.portfolio import Portfolio
from datetime import datetime


class TestPortfolio(unittest.TestCase): 
    def setUp(self):
        self.date_time = datetime(2021, 11, 5, 15, 30, 20)
        self.portfolio = Portfolio(10000)

              
    def test_on_fill_long(self):
        """
        Test long and sell multiple shares of GOOG at various price
        to check the arithmetic and cost handling
        """   
        goog = self.portfolio.fin_instruments
        #buy 5 shares of GOOG
        fillevent = FillEvent(self.date_time, 'GOOG', Direction.LONG, 'NASDAQ', 2000, 5, 1)
        self.portfolio.on_fill(fillevent)
        #sell 2 shares of GOOG
        fillevent = FillEvent(self.date_time, 'GOOG', Direction.SHORT, 'NASDAQ', 3000, 2, 1)
        self.portfolio.on_fill(fillevent)  
        #buy 1 shares of GOOL
        fillevent = FillEvent(self.date_time, 'GOOG', Direction.LONG, 'NASDAQ', 2500, 1, 1)
        self.portfolio.on_fill(fillevent)     
        #sell 3 shares of GOOL
        fillevent = FillEvent(self.date_time, 'GOOG', Direction.SHORT, 'NASDAQ', 3500, 3, 1)
        self.portfolio.on_fill(fillevent)
        
        self.assertEqual(self.portfolio.cash, 13996)
        self.assertEqual(self.portfolio.market_value, 3500)
        self.assertEqual(self.portfolio.get_asset_val(), 17496)
        self.assertEqual(goog.loc['GOOG']['POS'], 1)
        self.assertEqual(goog.loc['GOOG']['AVG_PRICE'], 2125)
        self.assertEqual(goog.loc['GOOG']['MKT_VAL'], 3500)
        self.assertEqual(goog.loc['GOOG']['RLZD_P'], 6125)
        self.assertEqual(goog.loc['GOOG']['UNRLZD_P'], 1375)

         
    def test_on_fill_short(self):
        """
        Test short and buy back multiple shares of GOOG at various price
        to check the arithmetic
        """
        goog = self.portfolio.fin_instruments
        #short 5 shares of GOOG
        fillevent = FillEvent(self.date_time, 'GOOG', Direction.SHORT, 'NASDAQ', 2000, 5, 1)
        self.portfolio.on_fill(fillevent)  
        #buy back 2 shares of GOOG
        fillevent = FillEvent(self.date_time, 'GOOG', Direction.LONG, 'NASDAQ', 1000, 2, 1)
        self.portfolio.on_fill(fillevent) 
        #short 2 shares of GOOL
        fillevent = FillEvent(self.date_time, 'GOOG', Direction.SHORT, 'NASDAQ', 2500, 2, 1)
        self.portfolio.on_fill(fillevent)
        #buy back 3 shares of GOOL
        fillevent = FillEvent(self.date_time, 'GOOG', Direction.LONG, 'NASDAQ', 3500, 3, 1)
        self.portfolio.on_fill(fillevent)
        
        self.assertEqual(self.portfolio.cash, 12496)
        self.assertEqual(self.portfolio.market_value, -7000)
        self.assertEqual(self.portfolio.get_asset_val(), 5496)     
        self.assertEqual(goog.loc['GOOG']['POS'], -2)
        self.assertEqual(goog.loc['GOOG']['AVG_PRICE'], 2200)
        self.assertEqual(goog.loc['GOOG']['MKT_VAL'], -7000)
        self.assertEqual(goog.loc['GOOG']['RLZD_P'], -1900)
        self.assertEqual(goog.loc['GOOG']['UNRLZD_P'], -2600)

        
    def test_on_bar_long(self):
        """
        test if the information updated to a stock is correct
        when the stock is in a long position
        """
        barevent = BarEvent(self.date_time, 'GOOG', 1000, 2500, 900, 2300, 100000)
        fillevent = FillEvent(self.date_time, 'GOOG', Direction.LONG, 'NASDAQ', 2000, 5, 1)
        goog = self.portfolio.fin_instruments
        self.portfolio.on_fill(fillevent)
        self.portfolio.on_bar(barevent)
            
        barevent = BarEvent(self.date_time, 'GOOG', 1000, 2000, 900, 1500, 100000)
        self.portfolio.on_bar(barevent)
        
        self.assertEqual(goog.loc['GOOG']['POS'], 5)
        self.assertEqual(goog.loc['GOOG']['AVG_PRICE'], 2000)
        self.assertEqual(goog.loc['GOOG']['MKT_VAL'], 7500)
        self.assertEqual(goog.loc['GOOG']['RLZD_P'], 0)
        self.assertEqual(goog.loc['GOOG']['UNRLZD_P'], -2500)    
        self.assertEqual(self.portfolio.cash, -1)
        self.assertEqual(self.portfolio.market_value, 7500)
        self.assertEqual(self.portfolio.get_asset_val(), 7499)
     
        
    def test_on_bar_short(self):
        """
        test if the information updated to a stock is correct
        when the stock is in a short position
        """
        barevent = BarEvent(self.date_time, 'GOOG', 1000, 2500, 900, 2300, 100000)
        goog = self.portfolio.fin_instruments
        fillevent = FillEvent(self.date_time, 'GOOG', Direction.SHORT, 'NASDAQ', 2000, 4, 1)
        self.portfolio.on_fill(fillevent)
        self.portfolio.on_bar(barevent)
        
        barevent = BarEvent(self.date_time, 'GOOG', 1000, 2500, 900, 3000, 100000)
        self.portfolio.on_bar(barevent)
        
        self.assertEqual(goog.loc['GOOG']['POS'], -4)
        self.assertEqual(goog.loc['GOOG']['AVG_PRICE'], 2000)
        self.assertEqual(goog.loc['GOOG']['MKT_VAL'], -12000)
        self.assertEqual(goog.loc['GOOG']['RLZD_P'], 0)
        self.assertEqual(goog.loc['GOOG']['UNRLZD_P'], -4000)    
        self.assertEqual(self.portfolio.cash, 17999)
        self.assertEqual(self.portfolio.market_value, -12000)
        self.assertEqual(self.portfolio.get_asset_val(), 5999)
               

        
if __name__ == '__main__':
    unittest.main(verbosity=2)
        

                

        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Dec 17 12:46:25 2021

@author: ron
"""
from .base import AbstractBroker
from ..event.event import Direction
from ..ibtws.orders import Orders
from ..ibtws.contracts import create_contract
import queue
import time


class IBBroker(AbstractBroker):
    
    def __init__(self, twsclient, event_queue, ticker_list, db_client=None):
        self.db_client = db_client
        self.event_queue = event_queue
        self.twsclient = twsclient
        self.contracts_list = {}
        for ticker in ticker_list:
            self.contracts_list[ticker] = create_contract(ticker)
  
     
    def execute_order(self, order_event):
        if time.time() - order_event.timestamp >= 300:
            print("Order rejected as the signal generated is at least 5 mins before")
        
        else:
            self.twsclient.reqIds()
            if order_event.direction == Direction.LONG:
                order = Orders.MarketOrder("BUY", order_event.quantity)

            #order_event.direction == Direction.SHORT    
            else:
                order = Orders.MarketOrder("SELL", order_event.quantity)
                
            self.twsclient.placeOrder(self.twsclient.nextValidOrderId, self.contracts_list[order_event.ticker], order)


    def get_fill_event(self):
        try:
            fill_event = self.twsclient.execution_detail.get(False)
        except queue.Empty:
            pass
        else:
            print(fill_event)
            self.event_queue.put(fill_event)
            if self.db_client is not None:
                self.save_to_db(fill_event)        
       



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Dec  7 20:02:36 2021

@author: ron
"""

from .twswrapper import TWSWrapper
from ..ibapi.client import EClient
from .contracts import create_contract
from ..event.event import BarEvent, FillEvent
import threading
from time import sleep
import queue
from copy import deepcopy


"""
The main thread will exit whenever it is finished 
executing all the code in your script 
that is not started in a separate thread.
"""
class TWSClient(TWSWrapper, EClient):

    exchange = "SMART"
    
    def __init__(self, ip, port_num, client_id):
        TWSWrapper.__init__(self)
        EClient.__init__(self, wrapper=self)
        self.fill_event =  FillEvent(None, None, None, None, None, None, None)
        #for storing the fill_event, then ib_broker will get the FIllEvent in the queue  
        self.execution_detail = queue.Queue()
        self.nextValidOrderId = 0
        #store the error code raised by the api, the main program
        #will get it and perform subseqent actions
        self.error_code = queue.Queue()
        self.cond = threading.Condition()
        self.time = None
        self.ip = ip
        self.port_num = port_num
        self.client_id = client_id
        self.end_time = None


    def connect(self):    
        print("Connecting to tws...")
        super().connect(self.ip, self.port_num, self.client_id)
        while not self.isConnected():
            pass

 
    def run(self):
        try: 
            if self.isConnected():
                print("Running tws client...")            
                #create thread for running the EClient message loop
                t1 = threading.Thread(target=super().run, name="EClient.run Thread")                                  
                t1.start()   
                sleep(1)     
        except:
            raise Exception("TWS client is not connected to tws yet!")
            
            
    def reqHistoricalData(self, reqId , contract, endDateTime,
                          durationStr, barSizeSetting, whatToShow="TRADES",
                          useRTH=1, formatDate=1, keepUpToDate=False):
        super().reqHistoricalData(reqId, contract, endDateTime,
                                 durationStr, barSizeSetting, whatToShow,
                                 useRTH, formatDate, keepUpToDate, [])  
        #time.sleep(1)


    def realtime_config(self, ticker_list, timeframe, whatToShow, useRTH):
        self.ticker_list = ticker_list
        self.timeframe = timeframe
        self._bars = {k: BarEvent(None, ticker_list[k], None, -1, 999999, None, 0 ) for k in range(0, len(ticker_list))}
        self.lastest_bar_event = queue.Queue()
        self.contracts_list = []
        self.whatToShow = whatToShow
        self.useRTH = useRTH
        #ib return starting time of the bar
        self.last_bar_time = 0 
        self.realtime_subscribed = False
        for ticker in ticker_list:
            self.contracts_list.append(create_contract(ticker))          


    def reqRealTimeBars(self):
        #reqRealTimeBars returns bar in every 5 seconds
        #these 5 second bars are then constructed into bar with defined timeframe by calling _construct_bar()
        #then it puts it into latest_bar_event queue which Data Handler can get the latest Bar event
        with self.cond:
            print("Wait until first real time bar return...")
            for i in range(len(self.contracts_list)):
                super().reqRealTimeBars(i, self.contracts_list[i], 5, self.whatToShow, self.useRTH, [])
            self.cond.wait()
        print("Data Subscription is successful!")  


    def resubscribe_RealTimeBars(self):
        """
        Since reqRealTimeBars may stop working 
        after IB server reset or between trading sessions
        it is necessary to resubscribe the real time bars       
        """
        #cancel exisiting subscription and resubscribe again
        for i in range(0, len(self.contracts_list)):
            super().reqRealTimeBars(i, self.contracts_list[i], 5, self.whatToShow, self.useRTH, [])        

    
    def cancel_subscription(self):
        print("cancelling existing subscribtion...")
        for i in range(len(self.contracts_list)):
            super().cancelRealTimeBars(i)
        sleep(1)
        self.realtime_subscribed = False
        print("Data Subscription cancelled successfully")

    
    def reqCurrentTime(self):
        with self.cond:
            super().reqCurrentTime()
            self.cond.wait()


    def reqHeadTimeStamp(self, reqId, contract, whatToShow, useRTH, formatDate):
        """returns earliest available data of a type of data for a particular contract"""
        super().reqHeadTimeStamp(reqId, contract, whatToShow, useRTH, formatDate)
            

    def reqHistoricalData(self, duration_str):
        with self.cond:
            for i in range(len(self.contracts_list)):   
                super().reqHistoricalData(i, self.contracts_list[i], "", duration_str, 
                                        "5 secs", self.whatToShow, 
                                        self.useRTH, 1, False, [])
            self.cond.wait()
            print("Missing data filled...")      

    
    def reqIds(self):
        with self.cond:
            super().reqIds(-1)
            self.cond.wait()
    

    def _construct_bar(self, ticker_id, current_bar):
        #constrtuct bar from 5 seconds bar in defined timeframe
        bar = self._bars[ticker_id]
        if bar.open_price == None and bar.timestamp == None:
            bar.open_price = current_bar.open_price
            bar.timestamp = current_bar.start_time
            # print("start time:", bar.timestamp)

        bar.high_price = max(bar.high_price, current_bar.high_price)
        bar.low_price = min(bar.low_price, current_bar.low_price)
        
        if current_bar.end_time != self.last_bar_time:           
            bar.volume += current_bar.volume
            
        #5 seconds bars are enough to create a bar with defined timeframe
        if (current_bar.end_time - bar.timestamp) >= self.timeframe or current_bar.end_time == self.end_time:
            # if(current_bar.end_time - bar.timestamp) > self.timeframe:
            #     print("there are 5sec bars missing in this bar")                
            bar.close_price = current_bar.close_price        
            self.lastest_bar_event.put(deepcopy(bar))
            #init next bar            
            bar.timestamp = None            
            bar.open_price = None
            bar.high_price = -1
            bar.low_price = 999999
            bar.close_price = None
            bar.volume = 0   
                                 
        self.last_bar_time = current_bar.end_time
            

     


from datetime import datetime, timedelta
import queue
import unittest
from tradingsystem.ibtws.twsclient import TWSClient
from tradingsystem.event.event import BarEvent
from tradingsystem.common import Bar
from tradingsystem.ibapi.common import BarData


def construct_BarData(time, open, high, low, close, volume, wap, count):
    bar = BarData()
    bar.date = time
    bar.open = open
    bar.high = high
    bar.low = low
    bar.close = close
    bar.volume = volume
    bar.wap = wap
    bar.barCount = count
    return bar


class TestTWSClient(unittest.TestCase):
    #test the fuctions in twsclient and twswrapper
    def setUp(self):
        self.lastest_bar_event = queue.Queue()
        self.twsclient = TWSClient("127.0.0.1", 7497, 0)
        self.twsclient.last_bar_time = 0 
        self.twsclient._bars = {0: BarEvent(None, "AAPL", None, -1, 999999, None, 0 )}
        self.twsclient.lastest_bar_event = queue.Queue()
        self.twsclient.timeframe = 60
        self.twsclient.realtime_subscribed = False
        
        self.bar_stream = []
        self.bar_stream.append(Bar(1644662475, 1644662480, 2, 4, 2, 3, 100, -1, -1))
        self.bar_stream.append(Bar(1644662480, 1644662485, 3, 5, 2, 4, 100, -1, -1)) 
        self.bar_stream.append(Bar(1644662485, 1644662490, 4, 5, 3, 3, 100, -1, -1)) 
        self.bar_stream.append(Bar(1644662490, 1644662495, 3, 4, 3, 2, 100, -1, -1)) 
        self.bar_stream.append(Bar(1644662495, 1644662500, 2, 7, 2, 7, 100, -1, -1)) 
        self.bar_stream.append(Bar(1644662500, 1644662505, 7, 10, 7, 9, 100, -1, -1)) 
        self.bar_stream.append(Bar(1644662505, 1644662510, 9, 13, 1, 12, 100, -1, -1)) 
        self.bar_stream.append(Bar(1644662510, 1644662515, 12, 15, 12, 14, 100, -1, -1)) 
        self.bar_stream.append(Bar(1644662515, 1644662520, 14, 14, 9, 10, 100, -1, -1)) 
        self.bar_stream.append(Bar(1644662520, 1644662525, 10, 11, 8, 9, 100, -1, -1)) 
        self.bar_stream.append(Bar(1644662525, 1644662530, 11, 11, 11, 11, 100, -1, -1))
        self.bar_stream.append(Bar(1644662530, 1644662535, 11, 12, 5, 10, 100, -1, -1))


    def test_construct_bar(self):
        #test whether a 1min candle bar can be correctly
        #constructed by streaming 5 sec candle bars
        aapl_bar = self.twsclient._bars[0]
        first_bar_time = 1644662475
        volume = 0
        max_price = float('-inf')
        min_price = float('inf')
        open_price = self.bar_stream[0].open_price
        
        for i in range(len(self.bar_stream)):
            self.twsclient._construct_bar(0, self.bar_stream[i])
            max_price = max(self.bar_stream[i].high_price, max_price)
            min_price = min(self.bar_stream[i].low_price, min_price)                              
            volume += self.bar_stream[i].volume
            
            if i != len(self.bar_stream) - 1:
                #candle bar expected at each 5s bar coming in
                expected_bar = BarEvent(first_bar_time, "AAPL", open_price, max_price, min_price, None, volume)
                self.assertEqual(aapl_bar, expected_bar)
            #last bar, fill the close price
            else:
                constructed_bar = self.twsclient.lastest_bar_event.get()
                expected_bar = BarEvent(first_bar_time, "AAPL", 2, 15, 1, 10, 1200)
                self.assertEqual(constructed_bar, expected_bar)
                
            self.assertEqual(self.twsclient.last_bar_time, self.bar_stream[i].end_time)
        
        #test whether the next new 1min bar can be constructed correctly
        #after receiving the 5 sec bar
        bar = Bar(1644662535, 1644662540, 10, 15, 10, 11, 100, -1, -1)       
        self.twsclient._construct_bar(0, bar)
        expected_bar = BarEvent(1644662535, "AAPL", 10, 15, 10, None, 100)
        self.assertEqual(aapl_bar, expected_bar)
        self.assertEqual(self.twsclient.last_bar_time, 1644662540)
        
       
    def test_construct_bar_missing_end(self):
        #test whether it can function normally 
        # if there is a 5 sec bar missing at the end
        aapl_bar = self.twsclient._bars[0]
        first_bar_time = 1644662475
        #pop out the last bar to simulate the last 5sec bar is missing
        self.bar_stream.pop()
        #the next 5sec bar received is the start of the next 1min bar
        self.bar_stream.append(Bar(1644662535, 1644662540, 11, 20, 5, 9, 100, -1, -1))
        volume = 0
        max_price = float('-inf')
        min_price = float('inf')
        open_price = self.bar_stream[0].open_price
        
        for i in range(len(self.bar_stream)):
            self.twsclient._construct_bar(0, self.bar_stream[i])
            max_price = max(self.bar_stream[i].high_price, max_price)
            min_price = min(self.bar_stream[i].low_price, min_price)                              
            volume += self.bar_stream[i].volume
            
            if i != len(self.bar_stream) - 1:
                expected_bar = BarEvent(first_bar_time, "AAPL", open_price, max_price, min_price, None, volume)
                self.assertEqual(aapl_bar, expected_bar)
            #last bar, fill the close price
            else:
                constructed_bar = self.twsclient.lastest_bar_event.get()
                expected_bar = BarEvent(first_bar_time, "AAPL", 2, 20, 1, 9, 1200)
                self.assertEqual(constructed_bar, expected_bar)
                
            self.assertEqual(self.twsclient.last_bar_time, self.bar_stream[i].end_time)
            
        #test whether the next new 1min bar can be constructed correctly
        #after receiving the 5 sec bar    
        bar = Bar(1644662540, 1644662545, 10, 15, 10, 11, 100, -1, -1)       
        self.twsclient._construct_bar(0, bar)
        expected_bar = BarEvent(1644662540, "AAPL", 10, 15, 10, None, 100)
        self.assertEqual(aapl_bar, expected_bar)
        

    def test_construct_bar_missing_middle(self):
        #test whether it can function normally 
        # if there are some 5 sec bars missing during the construction
        aapl_bar = self.twsclient._bars[0]
        first_bar_time = 1644662475
        #missming middle
        #remove Bar(1644662505, 1644662510, 9, 13, 1, 12, 100, -1, -1)
        del self.bar_stream[6]
        volume = 0
        max_price = float('-inf')
        min_price = float('inf')
        open_price = self.bar_stream[0].open_price       
        
        for i in range(len(self.bar_stream)):
            self.twsclient._construct_bar(0, self.bar_stream[i])
            max_price = max(self.bar_stream[i].high_price, max_price)
            min_price = min(self.bar_stream[i].low_price, min_price)                              
            volume += self.bar_stream[i].volume
            
            if i != len(self.bar_stream) - 1:
                expected_bar = BarEvent(first_bar_time, "AAPL", open_price, max_price, min_price, None, volume)
                self.assertEqual(aapl_bar, expected_bar)
            #last bar, fill the close price
            else:
                constructed_bar = self.twsclient.lastest_bar_event.get()
                expected_bar = BarEvent(first_bar_time, "AAPL", 2, 15, 2, 10, 1100)
                self.assertEqual(constructed_bar, expected_bar)
                
            self.assertEqual(self.twsclient.last_bar_time, self.bar_stream[i].end_time)


    def test_historicalData(self):
        #test whether the missing bar can be retrived
        #and be construced sucessfully when there is a network interruption
        #assume network interruption at 1644662537
        #and network restore at 1644662600 
        #then it gets 5 sec bars from 1644662535 to 1644662600
        aapl_bar = self.twsclient._bars[0]
        for bar in self.bar_stream:
            self.twsclient._construct_bar(0, bar)

        self.twsclient.lastest_bar_event.get()
        #network interruption at 1644662537 and restore at 1644662600
        backfill_bars = []
        backfill_time = []
        start_time = datetime.strptime("20220212 10:42:15", "%Y%m%d  %H:%M:%S")
        
        for _ in range(13):
            backfill_time.append(start_time.strftime("%Y%m%d %H:%M:%S"))
            start_time += timedelta(seconds=5)
             
        backfill_bars.append(construct_BarData(backfill_time[0], 10, 15, 9, 9, 100, -1, -1)) #20220212 10:42:15 (1644662535)
        backfill_bars.append(construct_BarData(backfill_time[1], 9, 10, 9, 9, 100, -1, -1)) #20220212 10:42:20 (1644662540)
        backfill_bars.append(construct_BarData(backfill_time[2], 9, 20, 15, 18, 100, -1, -1)) #20220212 10:42:25 (1644662545)
        backfill_bars.append(construct_BarData(backfill_time[3], 18, 20, 18, 20, 100, -1, -1)) #20220212 10:42:30 (1644662550)
        backfill_bars.append(construct_BarData(backfill_time[4], 20, 25, 19, 23, 100, -1, -1)) #20220212 10:42:35 (1644662555)
        backfill_bars.append(construct_BarData(backfill_time[5], 23, 23, 23, 23, 100, -1, -1)) #20220212 10:42:40 (1644662560)
        backfill_bars.append(construct_BarData(backfill_time[6], 23, 23, 18, 22, 100, -1, -1)) #20220212 10:42:45 (1644662565)
        backfill_bars.append(construct_BarData(backfill_time[7], 22, 30, 3, 25, 100, -1, -1)) #20220212 10:42:50 (1644662570)
        backfill_bars.append(construct_BarData(backfill_time[8], 25, 25, 15, 17, 100, -1, -1)) #20220212 10:42:55 (1644662575)
        backfill_bars.append(construct_BarData(backfill_time[9], 17, 18, 15, 15, 100, -1, -1)) #20220212 10:43:00 (1644662580)
        backfill_bars.append(construct_BarData(backfill_time[10], 15, 19, 10, 12, 100, -1, -1)) #20220212 10:43:05 (1644662585)
        backfill_bars.append(construct_BarData(backfill_time[11], 12, 15, 12, 11, 100, -1, -1)) #20220212 10:43:10 (1644662590)
        backfill_bars.append(construct_BarData(backfill_time[12], 11, 18, 14, 17, 100, -1, -1)) #20220212 10:43:15 (1644662595)

        for i in range(len(backfill_bars)):
            self.twsclient.historicalData(0, backfill_bars[i])   
        constructed_bar = self.twsclient.lastest_bar_event.get()
        
        expected_bar = BarEvent(1644662535, "AAPL", 10, 30, 3, 11, 1200)
        self.assertEqual(constructed_bar, expected_bar)
        expected_bar = BarEvent(1644662595, "AAPL", 11, 18, 14, None, 100)
        self.assertEqual(aapl_bar, expected_bar)
     

    def test_construct_bar_end(self):
        """
        test if the bar is constructed when market ends.
        """
        self.twsclient.end_time = 1644662530
        self.twsclient.lastest_bar_event = queue.Queue()
        self.bar_stream.pop()
        for i in range(len(self.bar_stream)):
            self.twsclient._construct_bar(0, self.bar_stream[i])

        self.assertEqual(self.twsclient.lastest_bar_event.qsize(), 1)

        


if __name__ == '__main__':
    unittest.main(verbosity=2)  


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Dec 14 14:43:25 2021

@author: ron
"""
from .base import AbstractDataHandler
import queue
import time
from ..common import check_ping



class IBRealTimeBarHandler(AbstractDataHandler):
    """
    Get real time bar data using tws api
    Parameters
    ----------
    twsclient : an instance of TWSClient
    timeframe: trading timeframe, it needs to be multiple of 5
    You may refer to https://interactivebrokers.github.io/tws-api/historical_bars.html
    for valid input of whatToShow and useRTH
    """
    def __init__(self, twsclient, ticker_list, timeframe, 
                 whatToShow, useRTH, event_queue, db_client=None
        ):
        self.db_client = db_client
        self.twsclient = twsclient
        self.event_queue = event_queue
        self.ticker_list = ticker_list
        self.timeframe = timeframe
        self.whatToShow = whatToShow
        self.useRTH = useRTH
        self.bars_store = []
        self.twsclient.realtime_config(self.ticker_list, self.timeframe, 
                                       self.whatToShow, self.useRTH)

                
    def get_next(self):
        try:
            bar_event = self.twsclient.lastest_bar_event.get(False)
        except queue.Empty:
            pass
        else:
            #convert to readable time
            #print("putting bar event")
            print(bar_event)
            self.event_queue.put(bar_event)
            self.bars_store.append(bar_event)
            if self.db_client is not None:
                self.save_to_db(bar_event)

                                   
    def request_data(self):
        #call this function first before calling get_next()
        self.twsclient.reqRealTimeBars()
        # while not self.twsclient.realtime_subscribed:
        #     pass
                                     
   
    def cancel_subscription(self):
        if self.twsclient.realtime_subscribed:
            self.twsclient.cancel_subscription()
        
        
    def check_bar_interruption(self):
        #check if bars arrive in expected time
        #if data is not arrived within 5mins, resubscribe again                    
            current_time = time.time()                 
            if current_time - self.twsclient.last_bar_time >= 300:
                print("Data isn't arrived in expected time, backfill missing data and resubscribe...")                               
                #ensure network is connected
                print("testing network connection...")
                if check_ping():
                    print("network is ok!")
                else:
                    raise Exception("network is not connected...")
                
                print("check if it is connected to IB...")
                if not self.twsclient.isConnected():
                    print("tws is not connected, reconnect again...")
                    self.twsclient.connect()
                    print("reconnection is successful!")
                    self.twsclient.run()

                print("tws is connected...")
                self.cancel_subscription()
                                    
                print("backfill missing data....")
                self.resubscribe()
                self.request_data()             


    def resubscribe(self):
        """
        after a historical data farm is disconnected,
        wait IB to connect it again.
        backfill the bars missing during the disconnected period,
        then request real time bar again
        """
        self.twsclient.reqCurrentTime()
        end_time = self.twsclient.time - (self.twsclient.time % 5)
        time_missing = end_time - self.twsclient.last_bar_time
        if time_missing < 60:
            duration_str = "60 S"
        else:
            duration_str = str(time_missing) + " S"
            
        self.twsclient.reqHistoricalData(duration_str)

            

      


"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""


"""
This module has tools for implementing the IB low level messaging.
"""


import struct
import logging
import sys

from .common import UNSET_INTEGER, UNSET_DOUBLE
from .utils import ClientException
from .utils import isAsciiPrintable
from .errors import INVALID_SYMBOL

logger = logging.getLogger(__name__)


def make_msg(text) -> bytes:
    """ adds the length prefix """

    msg = struct.pack("!I%ds" % len(text), len(text), str.encode(text))
    return msg


def make_field(val) -> str:
    """ adds the NULL string terminator """
    if val is None:
        raise ValueError("Cannot send None to TWS")

    # if string is not empty and contains invalid symbols
    if val is not None and type(val) == str and val and not isAsciiPrintable(val):
        raise ClientException(INVALID_SYMBOL.code(), INVALID_SYMBOL.msg(), val.encode(sys.stdout.encoding, errors='ignore').decode(sys.stdout.encoding))

    # bool type is encoded as int
    if type(val) is bool:
        val = int(val)

    field = str(val) + '\0'
    return field


def make_field_handle_empty(val) -> str:
    if val is None:
        raise ValueError("Cannot send None to TWS")

    if UNSET_INTEGER == val or UNSET_DOUBLE == val:
        val = ""

    return make_field(val)


def read_msg(buf:bytes) -> tuple:
    """ first the size prefix and then the corresponding msg payload """

    if len(buf) < 4:
        return (0, "", buf)
    size = struct.unpack("!I", buf[0:4])[0]
    logger.debug("read_msg: size: %d", size)
    if len(buf) - 4 >= size:
        text = struct.unpack("!%ds" % size, buf[4:4+size])[0]
        return (size, text, buf[4+size:])
    else:
        return (size, "", buf)


def read_fields(buf:bytes) -> tuple:
    if isinstance(buf, str):
        buf = buf.encode()

    """ msg payload is made of fields terminated/separated by NULL chars """
    fields = buf.split(b"\0")

    return tuple(fields[0:-1])   #last one is empty; this may slow dow things though, TODO






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Nov 13 19:29:40 2021

@author: ron
"""
from abc import ABC, abstractmethod


class AbstractTechnicalIndicator(ABC):
    """
    AbstractTechnicalIndicator is an abstract base class 
    for all technical indicator class
    """
    @abstractmethod
    def update(self, price):
        pass
    
    
    